// -*- mode: C++; c-basic-offset: 4; fill-column: 75; tab-width: 4; indent-tabs-mode: nil -*-
//  **** BEGIN LICENSE BLOCK ****
//  Copyright(c) 2005 Umesh Shankar and Chris Karlof
//  Based on LiveHTTPHeaders.js Copyright(c) 2002-2003 Daniel Savard and
//        on Generator.js Copyright(c) 2004  ABAS Software AG (Stefan Trcek)
//
//  Cookie Log:
//         This components allows replaying of a portion of a log while 
//         sending back on certain subsets of cookies received.
//         Eventually, it will log user events and replay those.
//
//  This program is free software; you can redistribute it and/or modify it under
//  the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
//  more details.
//
//  You should have received a copy of the GNU General Public License along with
//  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
//  Place, Suite 330, Boston, MA 02111-1307 USA
//  **** END LICENSE BLOCK ****


var oCookieLog;
var oCookiePermissions;
var mydocloader = Components.classes["@mozilla.org/docloaderservice;1"].getService(Components.interfaces.nsIWebProgress);

function startCookieLog() {
    oCookieLog = MakeCookieLog();
    oCookiePermissions = MakeCookiePermissions();
    oCookiePermissions.start();
    oCookieLog.start();
    oHeaderInfoLive.addObserver(oCookieLog);
    
    //dump_obj("oGenerator=", oGenerator);
    //dump("END startGenerator\n");

    mydocloader.addProgressListener(oCookieLog,Components.interfaces.nsIWebProgress.NOTIFY_ALL);

    cookieLogConsentInit();
  
    addObserver(oCookieLog, "cookie-changed", false);
    addObserver(oCookieLog, "cookie-rejected", false);
    addObserver(oCookieLog, "state-change", false);
    addObserver(oCookieLog, "trigger-domwindow", false);
    addObserver(oCookieLog, "trigger", false);
    addObserver(oCookieLog, "fixme", false);
    addObserver(oCookieLog, "sync-to-fork", false);
    addObserver(oCookieLog, "cookielog-set-debug", false);
    addObserver(oCookieLog, "domwindowclosed", false);
    addObserver(oCookieLog, "domwindowopened", false);

    addObserver(oCookieLog, NS_PREFBRANCH_PREFCHANGE_TOPIC_ID, false);
}

function stopCookieLog() {
    oCookiePermissions.savePermissions();
    oHeaderInfoLive.removeObserver(oCookieLog);
    mydocloader.removeProgressListener(oCookieLog);
    oCookieLog.stop();
    delete oCookieLog;
    oCookieLog = null;
    oCookiePermissions.stop();
    delete oCookiePermissions;
    oCookiePermissions = null;
}

function MakeCookieLog() {
    var cookielog = new HeaderInfoLive();
    cookielog.datapresent = "cookielog-datapresent";
    cookielog.lastClickedString = null;
    cookielog.lastClickedWindow = null;
    cookielog.lastClickedFormData = null;
    cookielog.isCapturing    = false;
    cookielog.COOKIE = 8;
    cookielog.START = 9;
    cookielog.STOP = 10;
    cookielog.REDIRECT = 11;
    cookielog.HOST = 12;
    cookielog.TRIGGER = 13;
    cookielog.REPLAY = 14;
    cookielog.TRIGGERFORM = 15;
    cookielog.INFO = 16;

    cookielog.debug = false;  
    cookielog.currentDocument = null;
    cookielog.isReplay = false;
    cookielog.userName = "";

    cookielog.nextWindowID = 1;
    cookielog.windowList = new Array;
  
    // these keep track of replay requests
    
    // The state machine that tracks ongoing replays; fields are 
    //   targetWin = XUL window in which we are replaying
    //   states = list of TRIGGER and START records we will replay
    //          = { index: index in original tree,
    //              type: TRIGGER or START,
    //              data: the trigger or start string }
    //   currentIndex = where we are in the replay list
    //   loadingURI = URI of page currently loading due to last trigger
    cookielog.replayState = null; 
    cookielog._replayStatusWindow = null;

    cookielog.showElementReload = true;
    cookielog.showElementClick = true;
    cookielog.showElementAdvance = true;

    cookielog.startPrefix = "START: ";
    cookielog.stopPrefix = "STOP: ";
    cookielog.triggerPrefix = "Triggered by ";
    cookielog.triggerFormPrefix = "Trigger form data: ";
    cookielog.acceptedCookiePrefix = "Accepted cookie: ";
    cookielog.rejectedCookiePrefix = "Rejected cookie: ";
    cookielog.appendedCookiePrefix = "Appending cookies to request: ";   

    cookielog.addRow = function(row, type) { 
        // Compute the style
        var oldrows = this.rows;
        var style;
        if (this.rows>0) {
            var ptype = this.type[this.rows-1];
            var pstyle = this.style[this.rows-1];
            if (type != ptype) {
                // We have a new type.  The line above may be single or last.
                if (pstyle == this.FIRST) {
                    pstyle = this.SINGLE;
                } else {
                    pstyle = this.LAST;
                }
                this.style[this.rows-1] = pstyle; // Update the style for the line above
                style = this.FIRST;
            } else {
                // Same type.  This line should be middle
                style = this.MID;
            }
        } else {
            // The first line.  No doubts, it should be of style FIRST.
            style = this.FIRST;
            document.getElementById(this.datapresent).removeAttribute('disabled');
        }
     
        // Add the row
        this.type.push(type);
        this.style.push(style);
        this.rows = this.data.push(row);
        if (row.length > this.hScrollMax) this.sethScroll(row.length);
        this.rowCountChanged(oldrows,(this.rows-oldrows));
    }
 
    cookielog.addSeparatorRow = function() {
        this.addRow("",this.SEPARATOR);  
    };

    cookielog.addStartRow = function(dom, url) {
        this.addRow("[" + this.windowList.indexOf(dom) + "] " + this.startPrefix + url, this.START);
    };

    cookielog.addStopRow = function(dom, url, title) {
        this.addRow("[" + this.windowList.indexOf(dom) + "] " + this.stopPrefix + url + " " + "TITLE=" + title, this.STOP);
    };

    cookielog.addRequestRow = function(dom, method, url) {
        this.addRow("[" + this.windowList.indexOf(dom) + "] " + method + " " + url, this.URL);     
    };

    cookielog.addTriggerRow = function(dom, clickString) {
        var winID = dom ? this.windowList.indexOf(dom) : 0;
        this.addRow("[" + winID + "] " + this.triggerPrefix + clickString, this.TRIGGER);
    };

    cookielog.addFormRow = function(dom, formData) {
        var winID = this.windowList.indexOf(dom);
        this.addRow("[" + winID + "] " + this.triggerFormPrefix + formData, this.TRIGGERFORM);
    };

    cookielog.addAcceptedCookieRow = function(cookieString) {
        this.addRow(this.acceptedCookiePrefix + cookieString, this.COOKIE);
    };

    cookielog.addRejectedCookieRow = function(cookieString) {
        this.addRow(this.rejectedCookiePrefix + cookieString, this.COOKIE);
        dumpToConsole("Cookie rejected: " + cookieString, "cookie-debug");
    };

    cookielog.addAppendedCookieRow = function(cookieString) {
        this.addRow(this.appendedCookiePrefix + cookieString, this.COOKIE);
      
    };

    // Parses form row into an array: [name1, value1, name2, value2...]
    // The values are still encoded with encodeURIComponent
    cookielog.parseFormValues = function(rowdata) {
        var formValues = new Array();
        var inputNames = new Array();
        rowdata = rowdata.replace(new RegExp("^\\[\\d+\\]\\s+"+oCookieLog.triggerFormPrefix), "");
        var values = rowdata.split(/&/);
        var tmp;
        var count = 0;
        for (i = 0; i < values.length; i++) {
            tmp = values[i].split(/=/);
            var index = inputNames.indexOf(tmp[0]);
            if(index < 0) {
                inputNames.push(tmp[0]);
                formValues[count*2] = tmp[0];
                formValues[count*2+1] = tmp[1];
                count++;
            }
        }
          
        return formValues;
    }


    // Parses START rows; returns the URL
    cookielog.parseStart = function(rowdata) { 
        if (! rowdata)
        return null;
        var tmp = rowdata.match(new RegExp("^\\[\\d+\\]\\s+" + this.startPrefix + "(\\S*)"));
        return tmp ? tmp[1] : null;
    }

    // Parses STOP rows; returns the URL and title in a two-element array
    cookielog.parseStop = function(rowdata) { 
        if (! rowdata)
        return null;
        var tmp = rowdata.match(new RegExp("^\\[\\d+\\]\\s+" + this.stopPrefix + "(\\S*) TITLE=(.*)"));
      
        return tmp ? tmp.slice(1) : null;
    } 



    cookielog.isReplaying = function() { return this.replayState != null; }
    cookielog.stopReplaying = function() { 
        if ( ! this.isReplaying() ) {
            return;
        }
        var win = this.replayState.targetWin;
        this.replayState = null; 

//         if ( win ) {
//             win.alert("Replay complete.")
//         }

        if ( this._replayStatusWindow ) {
            this._replayStatusWindow.close();
            this._replayStatusWindow = null;
        }
    }

  
    cookielog.observe = function(aSubject, aTopic, aData) {

        if (aTopic == 'trigger-domwindow') {
            this.lastClickedDOMWindow = aSubject;
        } else if (aTopic == 'trigger') {
            if (this.isReplaying()) {
                return;
            }

            if(aData) {

                //dumpToConsole("got trigger message with " + aData);
                this.lastClickedString = aData;
                this.lastClickedWindow = aSubject;
                this.lastClickedFormData = null;

                var forkOp = this.forkOpFromMainDOMWindow(this.lastClickedDOMWindow);
                var chromeLabel = null;
                if (! this.lastClickedString.match(/#document/) && ! this.lastClickedString.match(/chrome:/) ) {
		    
                    this.addTriggerRow(this.lastClickedDOMWindow, this.lastClickedString);
		    
                    var newTriggerIndex = this.rows - 1;
                    if (this.lastClickedWindow) {
                        var formData = saveAllFormData(this.lastClickedDOMWindow);
                        this.addFormRow(this.lastClickedDOMWindow, formData);
                    }        
                  
                    if ( forkOp && forkOp.getAppContent() ) {
                        var worked = forkOp.sendClick(this, newTriggerIndex);
                        if ( ! worked ) {
                            forkOp.suspectDifference();
                            return;
                        }
                    }

                } else if ( this.lastClickedString.match(/chrome:/) ) {

                    this.addTriggerRow(this.lastClickedDOMWindow, this.lastClickedString);		    
                  
                    chromeLabel = isAllowedChromeButton(this.lastClickedString);
                  
                    if ( chromeLabel && forkOp && 
                         forkOp.stillMirroring() ) {
                        var newTriggerIndex = this.rows - 1;

                        // Chrome clicks shouldn't count towards stepCount
                        if (forkOp.stepCount) {
                            forkOp.stepCount--;
                        }
                        if ( forkOp && forkOp.getAppContent() ) {
                            chromeButtonLabelToAction(forkOp.getForkWindow(), chromeLabel);
                        }
                    }
                }
            } else {
                //dumpToConsole("got NULL trigger message");
                this.lastClickedString = null;
                this.lastClickedWindow = null;
                this.lastClickedDOMWindow = null;                
            } 
//         } else if (aTopic == "sync-to-fork") {
//            var forkOp = this.forkOpFromMainWindow(aSubject);
//            forkOp.load(htmlWinFromXULWindow(aSubject).location.href);
//
//             var width=1920;
//             var height=1153;
//              var compareWin = openDialog("chrome://doppelganger/content/comparison.xul", 
//                                          "comparison",
//                                          "chrome,dialog=no,modal"
//                                          + ",left=" + ((width * 3/8)+ 2) + ",top=" + (height / 4) 
//                                          + ",outerWidth=" + ((width * 2/8) - 30) + ",outerHeight=350",
//                                         
//                                          function (arg) { doSwitch = arg; },
//                                          FP_SESSION,
//                                          "http://www.yahoo.com",
//                                          "policyref=\"http://p3p.yahoo.com/w3c/p3p.xml\", CP=\"CAO DSP COR CUR ADM DEV TAI PSA PSD IVAi IVDi CONi TELo OTPi OUR DELi SAMi OTRi UNRi PUBi IND PHY ONL UNI PUR FIN COM NAV INT DEM CNT STA POL HEA PRE GOV\""
//                                          );
       
        } else if (aTopic == 'fixme') {

            this.fixMe(aSubject);

        } else if (aTopic == 'state-change') {
	    
        } else if (aTopic == "cookie-changed" && aSubject != null) { // FIXME: Why is aSubject ever null?

            aSubject.QueryInterface(Components.interfaces.nsICookie2);
            dumpToConsole("Cookie event: " + aData + " " + aSubject.name + "=" + aSubject.value,"cookie-debug");
            if(aData != "deleted" && aData != "cleared") {
                this.addAcceptedCookieRow(cookieToString(aSubject));
            }

        } else if (aTopic == "cookie-rejected") {

            aSubject.QueryInterface(Components.interfaces.nsIURI);
            this.addRejectedCookieRow(aSubject.spec);

        } else if (aTopic == "cookielog-set-debug") {

            this.setDebug(aData != "false");

        } else if (aTopic == "domwindowopened") { // aSubject is actually a ChromeWindow, but .document is an HTMLDocument
//             // FIXME: open a new corresponding fork window
//             debugger;
//             if ( this.windowList.indexOf(aSubject) < 0 ) {
//                 this.windowList.push(aSubject);
//                 dumpToConsole("New DOM window opened; assigning ID " + (this.windowList.length - 1), "forkwindow-debug");
//                 this.nextWindowID++;
//             }

        } else if (aTopic == "domwindowclosed") { // aSubject is actually a ChromeWindow
            aSubject.QueryInterface(Components.interfaces.nsIDOMWindow);
            if ( ( typeof(mostRecentBrowserWindow) == "undefined" || ! mostRecentBrowserWindow() )
                 && ! window.closed ) {
                // all browser windows are gone: time to die
                /*
                try {
                    if ( getCookielogBranch().getBoolPref("userstudy") ) {
                        
                        var profd = getSpecialDir("ProfD");

//                         var dateStr = ((new Date()).toLocaleString()).replace(/[\s,l\\\'\"()]+/g, "-");
//                         profd.append("javascript-console-" + dateStr);
//                         //file.create(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0664);

                        profd.append((new Date().valueOf()));
                        profd.create(Components.interfaces.nsIFile.DIRECTORY_TYPE,0755);
                        
                        profd.append("javascript-console");
                        //file.create(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0664);
                        var consoleDumpStream = openForWriting(profd,false);                

                        if ( consoleDumpStream ) {
                            writeConsoleToFile(consoleDumpStream);
                        }

                        // write out the user's final session and persistent cookies
                        var data = "# Cookies\n";
                        var cmgr = Components.classes["@mozilla.org/cookiemanager;1"].getService();
                        cmgr = cmgr.QueryInterface(Components.interfaces.nsICookieManager);
                        var enumerator = cmgr.enumerator;
                        
                        while(enumerator.hasMoreElements()) {
                            var c = enumerator.getNext();
                            c.QueryInterface(Components.interfaces.nsICookie2);
                            data += cookieToString(c) + "\n";
                        }
                        
                        try {
                            profd = profd.parent;
                            profd.append("cookie-dump");
                            profd.create(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0664);
                            var foStream = openForWriting(profd,false);
                            foStream.write(data, data.length);
                            foStream.close();
                        } catch (e) {
                            alert("Cannot write out cookie state\n" + data);
                        }
                    }
                } catch(e) {
                }
                */
                window.close();
            }
        } else if (aTopic == NS_PREFBRANCH_PREFCHANGE_TOPIC_ID ) {

            if ( aData == "paranoia" ) {
                var Branch = getCookielogBranch();
                var newval = Branch.getCharPref("paranoia");
                if ( ! newval.match(/paranoia-(low|medium|high)/) ) {
                    Branch.setCharPref("paranoia", "paranoia-medium");
                } else {
                    this.paranoia = newval;
                }
            }

        }
    }
  
    // This function watches the HTTP response stream and populates the Cookie Log
    cookielog.observeResponse = function(oHttp, name, request, response, postData, isRedirect) {
        var cookies = response["Set-Cookie"];
        var redirect = response["Location"];
        var sentCookies = request["Cookie"];
        var reqArray = request["REQUEST"].split(" ");
        var host = request["Host"];
        var method = reqArray[0];
        var documentName = null;
        var documentNameSrc = null;
        var documentNameIndex = -1;
        var replayNameIndex = -1;
        var result = null;
        var progress = null;

        name = canonicalizeURL(name);
    
        var requestTailDomain = getTailURLDomain(getURLDomain(name));

        var documentDomain = null;

        var cookieManager = getCookieManager(); 
    
        // deal with redirects for replay
        // should be able to integrate this with new data structures
        var DOMWindow = getDOMWindowFromChannel(oHttp);
        var forkOp = this.forkOpFromDOMWindow(DOMWindow);

//         if(redirect) {
//             dumpToConsole("Redirect in observeReponse" + (forkOp ? " [FORK] " : "[MAIN]") + 
//                           ": changing\n" + name + " to\n" + redirect,"redirects-debug");
//         } 

        if( forkOp ) {
            //          dumpToConsole("Ignored observe response for fork window " + name, "forkwindow-debug");
            return;
        }


        if(cookies || sentCookies) {
            if(sentCookies) {
                this.addAppendedCookieRow(sentCookies);
            }
	
            this.addRequestRow(DOMWindow, method, name);

            this.addSeparatorRow();
            gCookiesWindow.refreshCookieDisplay();
        }
    };
  
    cookielog.observeGRequest = function (oHttp, url, method)
        {
            var DOMWindow = getDOMWindowFromChannel(oHttp);
            if ( ! DOMWindow ) {
                return;
            }

            var forkOp = this.forkOpFromDOMWindow(DOMWindow);
            if ( forkOp ) {
                forkOp.notifyRequest(oHttp, url, method);
                return;
            }

            if ( this.isReplaying() && method.toLowerCase() == "post" ) {
                this.replayState.madePostRequest(url, oHttp);
                return;
            }

            if ( this.paranoia == "paranoia-medium" && method.toLowerCase() == "post" ) {
                var tailDomain = getTailURLDomain(getURLDomain(url));
                
                if ( ! tailDomain ) { // e.g. "about:blank"
                    return;
                }

                var topTailDomain = getTailURLDomain(getURLDomain(DOMWindow.top.location.href));
                    
                if ( ! topTailDomain ) {
                    return;
                }

                var intMainPerms = oCookiePermissions.getDomainCookiePerms( tailDomain );
                if ( ! (intMainPerms & FP_SESSION) && 
                     (topTailDomain == tailDomain) ) { // suppress iframe ads, which have different domains from the top
                    dumpToConsole("Enabling session cookies for " + tailDomain + 
                                  " because a POST was seen with the medium paranoia setting");
                    oCookiePermissions.setDomainCookiePerms(tailDomain, FP_SESSION);
                }
            }            
        };

    
    cookielog.clearSiteAndVisitState = function() {
        forkGlobalState.clearVisitCounts();
        oCookiePermissions.clearAllPerms();
    }


    cookielog.forkOpFromDOMWindow = function(domWindow) {
        for(i in this.forkOperationsList) {
            if(domWindow == this.forkOperationsList[i].getDOMWindow()) {
                return this.forkOperationsList[i];
            }
        }
        return null;
    };

//     cookielog.forkOpFromMainWindow = function(mainWindow) {
//         for(i in this.forkOperationsList) {
//             if(mainWindow == this.forkOperationsList[i].getMainWindow()) {
//                 return this.forkOperationsList[i];
//             }
//         }
//         return null;
//     };

    cookielog.forkOpFromMainDOMWindow = function(domWindow) {
        for(i in this.forkOperationsList) {
            if(domWindow == this.forkOperationsList[i].getMainDOMWindow()) {
                return this.forkOperationsList[i];
            }
        }
        return null;
    };

    // mainWin is the XUL window corresponding to the primary (user-visible)
    // window which is to be mirrored
    cookielog.addNewForkWin = function(mainXUL, mainDOM, initialURL) {
        var height = this.debug ? 800 : 0;
        var width = this.debug ? 600 :  0;
        var forkWin = window.open("chrome://browser/content/browser.xul", "fork" + (new Date()).valueOf(), 
                                   "chrome,width="+width+",height="+height+
                                   ",menubar=no,toolbar=no,scrollbars=yes");
      
        if ( ! forkWin ) {
            alert("Critical Error: unable to create fork window!");
            return false;
        }

        var forkOperations = new ForkOperations(forkWin, mainDOM, this, oCookiePermissions);
        document.getElementById("fork-window-permissions").value = 
        parseCookiePermissionsToString(forkOperations.cookiePermissions);

        this.forkOperationsList.push( forkOperations );
        var fOL = this.forkOperationsList;
        
        // Just poll periodically and clean up the fork window; saves us having to deal
        // with multiple tabs' "unload" events, whatever they are
        var timer = window.setInterval( function() {
            if ( forkOperations.getMainDOMWindow() &&
                 forkOperations.getMainDOMWindow().closed ) {
                     fOL.splice(fOL.indexOf(forkOperations), 1);
                     forkWin.close();
                     window.clearInterval(timer);
                 }
        }, 2000);

        mainDOM.addEventListener("focus",
                                 function() {
            notifyObservers(mainDOM, "set-statusbar-sync", forkOperations.sync);
        }, true);

                        
        if ( ! oCookieLog.debug ) {
            hideDOMOrChromeWindow(forkWin);
            //                    forkWin.minimize();
            //                    forkWin.resizeTo(0,0);
        } else {
            unhideDOMOrChromeWindow(forkWin);
            //                    forkWin.resizeTo(800, 600);
        }


        var loadFn = function()  {
            setTimeout(function() {
                // Don't persist any of the fork window's traits
                forkWin.document.getElementById("main-window").setAttribute("persist", "");

                if ( forkWin.document && ! forkWin.forkBrowserOverlay ) {
                    forkWin.document.loadOverlay("chrome://doppelganger/content/forkBrowserOverlay.xul", null);
                }

                var firstURL = initialURL ? initialURL : mainDOM.location;
                dumpToConsole("Opening new fork window with URL " + firstURL, "forkwindow-debug");

                if ( ! forkOperations.load(firstURL) ) {
                    throw "fork window not ready";
                }

                forkWin.removeEventListener("load", loadFn, true);
            }, 0);
        };


        forkWin.addEventListener("load", loadFn, true);
    };

    cookielog.setDebug = function(aDebug) {
        if ( this.debug == aDebug ) {
            return;
        }

        this.debug = aDebug;

        for ( i in this.forkOperationsList ) {
            var forkOp  = this.forkOperationsList[i];
            if ( this.debug ) {
                if ( forkOp ) {
                    unhideDOMOrChromeWindow(forkOp.getForkWindow());
                    //                    forkOp.getForkWindow().restore();
                    //                    forkOp.getDOMWindow().resizeTo(800,600);
                }
                
                unhideDOMOrChromeWindow(window);
                //                window.restore();
                //                window.resizeTo(1024, Math.min(1000, screen.availHeight));
            } else {
                if ( forkOp ) {
                    //                    forkOp.getDOMWindow().resizeTo(0,0);
                    //                    forkOp.getForkWindow().minimize();
                    hideDOMOrChromeWindow(forkOp.getForkWindow());
                }
                
                hideDOMOrChromeWindow(window);
                //                window.resizeTo(0,0);            
                //                window.minimize();
            }
        }


        var Branch = getCookielogBranch();
        Branch.setBoolPref("debug", aDebug);
    };


    // Initialisation and termination functions
    cookielog.start = function() {
        this.oDump = document.getElementById("cookielog-dump");
        this.oDump.treeBoxObject.view = this;

        // The 'debug' pref corresponds to the checkbox on the statusbar
        // and controls whether the fork & cookielog window are seen
        var Branch = getCookielogBranch();
        
        if ( ! Branch.prefHasUserValue("debug") ) {
            Branch.setBoolPref("debug", false);
        }

        this.debug = Branch.getBoolPref("debug");
        var debug = this.debug;
        
        if ( ! debug ) {
            hideDOMOrChromeWindow(window);
            //                        window.resizeTo(0,0); 
                //                        window.minimize();
        } else {
            window.resizeTo(1024, Math.min(1000, screen.availHeight)); 
        }
 
            //             var resizeTimer = function() { 
            //                 window.setTimeout( function() { 
            //                     if ( ! debug ) {
            //                         hideDOMOrChromeWindow(window);
            //                         //                        window.resizeTo(0,0); 
            //                         //                        window.minimize();
            //                     } else {
            //                         window.resizeTo(1024, Math.min(1000, screen.availHeight)); 
            //                     }
            //                 }, 0 );
            //              };
            //window.addEventListener("load", resizeTimer, 0);
        //}

        if ( ! Branch.prefHasUserValue("paranoia") ) {
            Branch.setCharPref("paranoia", "paranoia-medium");            
        }

        this.paranoia = Branch.getCharPref("paranoia");
        Branch.addObserver("paranoia", this, false);

        //
        // Convert persistent cookies for sites for which we have not
        // decided about persistent cookies into fork cookies so we can
        // fork on them when the time comes; don't clobber existing fork
        // cookies, though.
        //
        var cookieManager = getCookieManager();
        var prefix = "FORK";
        var forkCookieList = new Array();
        cookies = cookieManager.enumerator;
        while (cookies.hasMoreElements()) {
            var cookie = cookies.getNext();
            cookie.QueryInterface(Components.interfaces.nsICookie2);
            forkCookieList[cookie.host+cookie.name+cookie.path] = 1;
        }



        cookies = cookieManager.enumerator;
        while (cookies.hasMoreElements()) {

            var cookie = cookies.getNext();
            cookie.QueryInterface(Components.interfaces.nsICookie2);
            if ( cookie.isSession ) {
                continue;
            } 

            var permSet = oCookiePermissions.getDomainCookiePerms(getTailURLDomain(cookie.host));

            if ( permSet & NO_FP_PERSISTENT || ! (permSet & FP_SESSION) ) {
               cookieManager.remove(cookie.host, cookie.name, cookie.path, false);
               continue;
            }

            if ( cookie.name.substr(0, prefix.length) != prefix &&
                 ! ( permSet & FP_PERSISTENT) ) 
                {
                    var newCookie = new Cookie(cookie);
                    newCookie.name = prefix + newCookie.name;
                    
                    if ( ! forkCookieList[newCookie.host+newCookie.name+newCookie.path] ) {
                        addCookieObject(newCookie);
                    }

                    cookieManager.remove(cookie.host, cookie.name, cookie.path, false);
                }

        }
        

        // Set scrollbar
        this.hScrollBar = document.getElementById("cookielog-dump-scroll");
        setInterval(this.hScrollHandler,100);
      
        this.initAtoms();

        this.forkOperationsList = new Array();

        var mainWin = mostRecentBrowserWindow();
        var mainDOM = htmlWinFromXULWindow(mostRecentBrowserWindow())
        this.windowList.push(null);
        this.windowList.push(mainDOM);
        this.addNewForkWin( mainWin, mainDOM, null );

        window.addEventListener("unload", 
                                function() 
        {
            for ( i in cookielog.forkOperationsList ) {
                cookielog.forkOperationsList[i].getForkWindow().close();
            }
        }, true);
        

 
        if( ! Branch.prefHasUserValue("forkwindow-debug") ) {
            Branch.setBoolPref("cookie-debug",false);
            Branch.setBoolPref("domwindow-debug",false);
            Branch.setBoolPref("userevent-debug",false);
            Branch.setBoolPref("formdata-debug",false);
            Branch.setBoolPref("domains-debug",false);
            Branch.setBoolPref("redirects-debug",false);
            Branch.setBoolPref("statechange-debug",false);
            Branch.setBoolPref("forkwindow-debug",false);
            Branch.setBoolPref("replay-debug",false);
        }

        if ( ! Branch.prefHasUserValue("userName") ) {
            this.userName = this.getUserInfo();
        } else {
            this.userName = Branch.getCharPref("userName");
        }
        
        this.makeUsernameRE();

        return true;
    };

    cookielog.getUserInfo = function() {            
        var Branch = getCookielogBranch();
        var username = "";
        
        var oldValue = Branch.prefHasUserValue("userName") ? Branch.getCharPref("userName") : "";

        window.openDialog("chrome://doppelganger/content/UserInfoDialog.xul", 
                          "username",
                          "chrome,dialog=yes,modal",
                          function(arg) { username = arg; },
                          oldValue
                          );
        
        Branch.setCharPref("userName", username);        
       
        return username;
    };

    cookielog.makeUsernameRE = function() {
        this.userNameRE = null;
        var names = cookielog.userName.match(/\S+/g);
        if ( names ) {
            for ( i = 0; i < names.length; i++ ) {
                names[i] = names[i].toLocaleLowerCase();
            }

            var reStr = "(\\b" + names[0] + "\\b)";
            for ( i = 1; i < names.length; i++ ) {
                reStr += "|(\\b" + names[i] + "\\b)";
            }

            this.userNameRE = new RegExp(reStr, "i");
        }

    };

    cookielog.hScrollHandler = function() {
        // Need to use global oGenerator object because 'this'
        // doesn't seem to be available.
        var base = oCookieLog;
        var curpos = base.hScrollBar.attributes.getNamedItem("curpos").value;
        if (curpos != base.hScrollPos) {
            base.hScrollPos = curpos;
            base.tree.invalidate();
        }
    };

    //   cookielog.replayURL = function() {
    //     var selection = this.oDump.view.selection;
    //     var index = selection.currentIndex;

    //     if(this.type[index]!=this.START) {
    //       alert("This is not the start of a request.");
    //       return;
    //     }
    
    //     // for now assume they are all GETs
    //     var tmp = this.data[index].match(/START: (.*)/);
    //     if(!tmp) {
    //       dumpToConsole("Error in replaying URL: " + this.data[index]);
    //       return;
    //     }      
    //     var url = tmp[1];
    
    //     var browser = window.opener.getWebNavigation();

    //     // Must clear the url we want to replay from the cache
    //     var classID = Components.classes["@mozilla.org/network/cache-service;1"];
    //     var cacheService = classID.getService(Components.interfaces.nsICacheService);
    
    //     try {
      
    //       var cacheSession = cacheService.createSession('HTTP',0,true);
    //       //var cacheSession = cacheService.createSession('HTTP-memory-only',1,true);
    //       var cacheEntryDescriptor = cacheSession.openCacheEntry(url.match(/^[^#]+/) ,
    // 							     Components.interfaces.nsICache.ACCESS_WRITE,true);
    //       cacheEntryDescriptor.doom();
    //       cacheEntryDescriptor.close();
    //     } catch (ex) {
    //       // Not able to clear disk cache selectivly, clear all of it!
    //       cacheService.evictEntries(Components.interfaces.nsICache.STORE_ON_DISK);
    //     }
    
    //     // Must clear memory cache because the above tips doesn't work for memory...
    //     cacheService.evictEntries(Components.interfaces.nsICache.STORE_IN_MEMORY);

    //     this.isReplay = true;
    //     this.currentDocument = url;
    //     browser.loadURI(url,0,null,null,null);

    //     // move selection to next START triggered by user event
    //     index++;
    //     do {
    //       index++;
    //     } while (index < this.rowCount && this.type[index] != this.TRIGGER);
    
    //     if ( index == this.rowCount ) {
    //       alert("Last START event.");
    //       return;
    //     }
    
    //     selection.select(index-1);
    //     this.tree.ensureRowIsVisible(index);
    
    //   };


    cookielog.onselect = function() {
        var selection = this.oDump.view.selection;
        var index = selection.currentIndex;

        if(this.type[index] == this.START) {
            var domain = getTailURLDomain(getURLDomain(this.parseStart(this.data[index])));
            document.getElementById("cookiepermissions-domain-input").value = domain;
        }
    };

    cookielog.setDebugPrefs = function(type) {
        var Branch = getCookielogBranch();
        var checkbox = document.getElementById(type);
        Branch.setBoolPref(type,checkbox.checked);
    };

    cookielog.setShowElementReload = function(doReload) { this.showElementReload = doReload; };
    cookielog.setShowElementClick = function(doClick) { this.showElementClick = doClick; };
    cookielog.setShowElementAdvance = function(doAdvance) { this.showElementAdvance = doAdvance; };

    cookielog.getCellText = function(row, column) {
        //return this.data[row].match(/^.*/);
        return this.data[row].substr(this.hScrollPos).match(/^.*/);
    };
  
    cookielog.advanceToNextTrigger = function() {      
        var selection = this.oDump.view.selection;
        var index = selection.currentIndex;
        var crossedStart = false;
        do {
            index++;
            if (this.type[index] == this.START)
                crossedStart = true;
        } while (index < this.rowCount && this.type[index] != this.TRIGGER);
      
        if ( index == this.rowCount ) {
            return crossedStart;
        }
      
        selection.select(index);
        this.tree.ensureRowIsVisible(index);
      
        return crossedStart;
    };

    // Given a trigger string, find the referenced element and highlight it with a border.
    // The trigger string is the currently-selected line in the Cookie Log dialog box
    // Also, fill in forms on the page, based on the form data string that follows the trigger.
    //
    // appcontent = appcontent element of the window to load into; if null, most recent one
    // triggerIndex = index in the tree of the trigger to use (-1 = selected one)
    // loadPage, sendClick, advanceSelection =
    //    Optionally (based on the arguments, or the dialog checkboxes as backup), 
    //    a) load the enclosing page (default: use what's there)
    //    b) click on that element
    //    c) move the selection to the next trigger string (not done if triggerIndex given)
    //
    // Return Values: 
    // Throws exception if an error occurs (for asynchronous errors after
    // page loads, no exception); otherwise a boolean, true iff
    // advanceSelection crossed a START event
    // 
    //
    cookielog.showElement = function(xulWindow, triggerIndex, loadPage, sendClick, advanceSelection) {
        var selection = this.oDump.view.selection;

        var index;
        if (triggerIndex >= 0) {
            index = triggerIndex;
            advanceSelection = false;
        } else {
            index = selection.currentIndex;
        }
    
        if (loadPage == null)
        loadPage = this.showElementReload;
      
        if (sendClick == null)
        sendClick = this.showElementClick;
      
        if (advanceSelection == null)
        advanceSelection = this.showElementAdvance;


        if(this.type[index] != this.TRIGGER) {
            throw new Error("This is not a trigger.");
        }
          
        var rowdata = this.data[index];
        var parts = parseTrigger(rowdata);

        var contentWindow = null;
        if (xulWindow == null) {
            xulWindow = mostRecentBrowserWindow();
        } 

        var contentWindow = parts["isChrome"] ? xulWindow : htmlWinFromXULWindow(xulWindow);

        var formValues;      
        if (this.type[index+1] == this.TRIGGERFORM) {
            var tmp = this.data[index+1];
            formValues = this.parseFormValues(tmp);
        }

        var highlightAndClick = function(arg_element) {
            arg_element.style.setProperty("border", "ridge", "important");
            arg_element.style.setProperty("border-width", "thick", "important");
            arg_element.style.setProperty("border-color", "yellow", "important");

            if (arg_element.scrollIntoView)
            arg_element.scrollIntoView();

            if(arg_element.scrollIntoView) {
                arg_element.scrollIntoView(false);
            }

            if (sendClick) {
                doClick(true, arg_element, parts["button"], 1, parts["clickX"], parts["clickY"]);
            }

            return 1;
        };

        var changeStyle = 
        function (arg_tagName, arg_idString, arg_window, arg_isChrome, formdata) {
            var f;
            var match = null;
          
            if (formdata && ! arg_isChrome) {
                tryFormFilling(contentWindow, formdata);
            }
          
            if ((match = findElement (contentWindow, arg_tagName, arg_idString))) {
                // Each match value is 0 if no match, 1 if both null, and 2 if matched
                if (match["exact"] || 
                    match["href"] > 1 ||
                    match["id"] > 1 ||
                    ( match["type"] > 1 && match["value"] ) ||
                    ( match["name"] && match["textContent"] ) ||
                    ( match["tagName"] && match["textContent"] ) ) {
                        highlightAndClick(match.elt);
                        return 1;
                    }
            }

            return 0;            
        };
 

        var windowLoad = function (aEvent) { 
            xulWindow.removeEventListener("load", windowLoad, true);
               
            var mainDoc = contentWindow.document;
            var doc = aEvent.originalTarget;
                     
            if (doc.URL != mainDoc.URL)
            return;
                     
            if (! changeStyle(parts["eltTagName"], parts["idString"], null, parts["isChrome"], formValues) ) {
                // FIXME: How to handle this error asynchronously?
                dumpToConsole("ERROR: Could not find the element specified in showElement: " + parts["idString"]);
            }
        };
   
        if (loadPage && ! parts["isChrome"]) {
            xulWindow.addEventListener("load", windowLoad, true);
            contentWindow.location = parts["iframe"] ? parts["mainURI"] : parts["uri"];
        } else if (! changeStyle(parts["eltTagName"], parts["idString"], null, parts["isChrome"], formValues) ) {
            throw new Error("Could not find element in showElement");
        }
    
        // Advance selection to the next trigger if desired
        if (advanceSelection) {
            return this.advanceToNextTrigger();
        } else {
            return false;
        }
    };


    cookielog.fixMe = function(win) {
            this.cookieReplay(win);
    };


    //
    // The implementation for the "Fixme" button
    // 
    // Phase I: Find where to start replaying
    //          Currently we go backwards in the history until the site changes
    //          Mark replay endpoint as the current page
    //
    // Phase II: Call showElement (load=yes, click=yes, advance=yes) (1st page)
    // 
    // Phase III: Call showElement (load=no, click=yes, advance=yes) (subsequent pages)
    //       ===> This phase is handled in onStateChange(), because we need to wait for the pages to load
    // 

    cookielog.cookieReplay = function (win) { 

        var domWindow = htmlWinFromXULWindow(win); 
        var thisWindowID = this.windowList.indexOf(domWindow);
        var tailDomain = getTailURLDomain(domWindow.location.hostname);

        //
        // If you click FIXME while still mirroring, then just show the
        // comparison window and see if that helps
        //
        var forkOp = this.forkOpFromMainDOMWindow(domWindow);
        if ( forkOp && forkOp.stillMirroring() ) {
            dumpToConsole("FIXME button pressed while still mirroring: invoking suspectDifference()", "replay-debug");
            forkOp.suspectDifference();
            return;
        }
        
        var perms = oCookiePermissions.getDomainCookiePerms(tailDomain);
        if ( perms & TP_SESSION || (forkGlobalState.doneDomains[tailDomain] == 1) ) {
            // we're out of options: max perms reached, or cookies already enabled this session
            dumpToConsole("FIXME button pressed, but max perms for this session already reached; perms: " + parseCookiePermissionsToString(perms), "replay-debug");
            win.alert("Unfortunately, this button cannot solve your problem because it's likely not cookie related.\n" +
                      "Please check the help section of the site you are visiting or contact the site operator.");
            return;
        }

        //
        // Set the right permission set: clear NO_*, add one more perm than before
        //
        var prompt = "Doppelganger is about to invoke its automated error recovery mechanism.\n" +
        "      It may result in changes to your privacy (cookie) policy\n" + 
        "      and a replaying of your session at this site.\n\n" +
        "If you do NOT wish to proceed, please click 'Cancel'.\nOtherwise, click 'OK'. ";        
        if ( ! win.confirm(prompt) ) {
            dumpToConsole("FIXME button pressed, but user did not agree to replay");
            return;
        }
        dumpToConsole("FIXME button pressed, replay initiated");

        // one more perm than before
        // FIXME: Check if this can possibly help, given the current cookie situation       
        dumpToConsole("[REPLAY] Old cookie perms for " + tailDomain + ": " + parseCookiePermissionsToString(perms));
        perms &= ~ (NO_FP_PERSISTENT | NO_FP_SESSION | NO_TP_SESSION | NO_TP_PERSISTENT);
        perms = ( perms * 2 ) + 1;
        oCookiePermissions.setDomainCookiePerms(tailDomain, perms);
        dumpToConsole("[REPLAY] New cookie perms for " + tailDomain + ": " + parseCookiePermissionsToString(perms));

        forkGlobalState.doneDomains[tailDomain] = 1;

        //
        // Now determine the list of things to replay
        // 

        var selection = this.oDump.view.selection;

        var replayStart;
        var replayEnd = this.rowCount - 1;
        if (replayEnd < 0) {
            win.alert("No browsing history available; can't replay.");
            return;
        }
        
//         if ( forkOp && forkOp.lastClickIdx != null ) { // mirroring *was* taking place
//             replayStart = forkOp.lastClickIdx; // pick up where we left off
//             forkOp.copyCookiesToMain(false, tailDomain);
//         } else 
        {
            replayStart = replayEnd;

            // Find the last trigger that is *not* from this site (not counting chrome)
            // Presumably it is the one that launched this site
            var i;

            for (i = replayEnd; i >= 0; i--) {
                if (this.type[i] == this.TRIGGER) {
                    if ( getWindowIDFromRow(this.data[i]) != thisWindowID ) {
                        continue;
                    }
                    
                    selection.select(i);

                    var trigger = parseTrigger(this.data[i]);
	      
                    if (trigger["uri"] == null) {
                        win.alert("Error: Could not get URL from " + this.data[i]);
                        return;
                    }

                    if ( ! isChromeURL(trigger["uri"]) &&
                         getTailURLDomain(getURLDomain(trigger["uri"])) != tailDomain )
                        break;
                }
            } 

            if ( i < 0 ) {
                i = 0;
            }
            
            replayStart = i;
        }


        // Now we advance to the next START or STOP event that matches the current site,
        // which we assume is the first page in this "visit" to the site
        for (; replayStart <= replayEnd; replayStart++) {
            if ( getWindowIDFromRow(this.data[replayStart]) != thisWindowID ) {
                continue;
            }
            
            if ( this.type[replayStart] == this.START ) {            
                var url = new URL(this.parseStart(this.data[replayStart]));
                if ( getTailURLDomain(url.getHost()) == tailDomain ) {
                    break;
                }
            } else if ( this.type[replayStart] == this.STOP ) {
                // The tricky part here is that sometimes you START on one site
                // and STOP on another e.g. gmail.com -> google.com/mail
                var url = new URL(this.parseStop(this.data[replayStart])[0]);
                if ( getTailURLDomain(url.getHost()) != tailDomain ) {
                    continue;
                }
                // now, rewind to the previous START
                while ( this.type[replayStart] != this.START ) {
                    replayStart--;
                }
                break;
            }
        }

        if (replayStart == replayEnd) {
            win.alert("Could not find anything to replay.");
            return;
        }

        selection.select(replayStart);

        var replayList = new Array();
        for (i = replayStart; i <= replayEnd; ++i) {
            if ( getWindowIDFromRow(this.data[i]) != thisWindowID ) {
                continue;
            }

            if ( this.type[i] == this.TRIGGER || this.type[i] == this.START 
                 || this.type[i] == this.STOP )
                replayList[replayList.length] = { 
                    index: i,
                    type: this.type[i],
                    data: this.data[i]
                };
        }

        this.replayState = new ReplayState(this, win, replayList);
        if ( this.replayState.sanityCheck() ) {
            this._replayStatusWindow = openStatusWindow(window,
                                                        "Replaying your session...",
                                                        "Doppelganger is replaying your session with cookies enabled to try to fix the site's functionality. Please wait until the replay is complete.",
                                                        win,
                                                        false,
                                                        null);
            this.replayState.go();
        } else {
            win.alert("Error: replay state did not pass initial check. Please check the log.");
            this.stopReplaying();
        }
    };
  
    cookielog.stateIsRequest = false;
  
    cookielog.QueryInterface = function(aIID) {
        if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
            aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
            aIID.equals(Components.interfaces.nsISupports)) {
                return this;
            }

        throw Components.results.NS_NOINTERFACE;
    };
  
    //
   // These member functions are to implement the WebProgressListener interface
    // Right now only onStateChange does anything
    //
  
    cookielog.onProgressChange = function(aProgress, aRequest, aCurSelfProgress, aMaxSelfProgress, aCurTotalProgress, aMaxTotalProgress) {
    };
      
    cookielog.onStatusChange = function(aProgress, aRequest, aStatus, aMessage) {
    };
  
    cookielog.onSecurityChange  = function(a,b,c){};
    cookielog.onLinkIconAvailable = function(a){};

    cookielog.parseStateFormat = function(aData) {
        if(aData == "START") {
            return this.START;
        } else if (aData == "STOP") {
            return this.STOP;
        } else if (aData == "REDIRECT") {
            return this.REDIRECT;
        } else if (aData == "REPLAY") {
            return this.REPLAY;
        } else
        return this.REQUEST;
    };

    cookielog.onStateChange = function(aProgress, aRequest, aFlag, aStatus) {
        try {
            if(!aRequest || !aRequest.name || (aRequest.name.indexOf("http") != 0)) {
                return 0;
            }
        } catch (ex) {
            return 0;
        }

        if ( ! ( (aFlag & STATE_IS_WINDOW) ||
                 ( (aFlag & STATE_REDIRECTING) && (aFlag & STATE_IS_DOCUMENT) ) ) ) { // apparently HTTP 302 redirects don't come with STATE_IS_WINDOW
            return 0;
        }

        aRequest.QueryInterface(Components.interfaces.nsIRequest);
        aProgress.QueryInterface(Components.interfaces.nsIWebProgress);
        var DOMWindow = aProgress.DOMWindow;
        var requestURL = canonicalizeURL(aRequest.name);
        //    var didLoadIntoFork = false;

        if ( DOMWindow.top != DOMWindow ) {
            return 0;
        }

        var forkOp = this.forkOpFromDOMWindow(DOMWindow);
        var win = forkOp ? "[FORK] " : "[MAIN] "
        dumpToConsole(win + aRequest.name + ": " + parseFlags(aFlag),"statechange-debug");
    
        if ( forkOp ) {            
            
            var visitor = {
                p3p: null,
                visitHeader: function(name, value) {
                    if ( name.toLowerCase() == "p3p" )
                        this.p3p = value;
                }        
            };

            if ( aFlag & STATE_STOP ) {
                var oHttp = aRequest.QueryInterface(Components.interfaces.nsIHttpChannel);
                oHttp.QueryInterface(Components.interfaces.nsIHttpChannelInternal);

                try {
                    var response = oHttp.visitResponseHeaders(visitor);
                }
                catch (e) {
                }
            }

            if ( aFlag & STATE_START ) {
                forkOp.setTopLevelRequest(requestURL, DOMWindow);
                forkOp.notifyStateChange(false, true, aStatus, requestURL, visitor.p3p);
            } else if ( aFlag & STATE_STOP ) {
                forkOp.notifyStateChange(false, false, aStatus, requestURL, visitor.p3p);
            }

            return 0;
        }

        // This is for the race condition where the fork window is open but
        // no ForkOperations object is attached to it yet
        if ( DOMWindow.name.match(/^fork/) ) {
            return 0;
        }

        var formData = null;
        //      var newTriggerIndex = -1;

        // FIXME: with REPLAYs, there is no current document currently set... need to figure out a strategy for this
        if ( (aFlag & STATE_START) ) { //&& (aFlag & STATE_IS_DOCUMENT) ) {

            // The idea here is that when we click on something because of a
            // trigger, we don't know what URL it will load, so we assume that the
            // first START event after a trigger is the one due to it.
            if ( this.isReplaying() && 
                 htmlWinFromXULWindow(this.replayState.targetWin) == DOMWindow && 
                 ! this.replayState.loadingURI ) {
                this.replayState.loadingURI = requestURL;
            }
      
//             // "Active windows" are those that were created by a load initated by a user action
//             // We maintain a list of these windows. Here we have a start event for a request,
//             // so we check to see if this document is being loaded in a active window
//             var found = ( this.activeWindows.indexOf(DOMWindow) >= 0 );
            
            if ( this.windowList.indexOf(DOMWindow) < 0 ) {
                // Kill popups spawned by fork windows
                if ( this.forkOpFromDOMWindow(DOMWindow.opener) ) {
                    dumpToConsole("Killed a fork window popup for url " + requestURL, "forkwindow-debug");
                    DOMWindow.close();
                    return 0;
                }

                this.windowList.push(DOMWindow);
                this.addNewForkWin(null, DOMWindow, requestURL);
                dumpToConsole("New DOM window opened at location " + requestURL + "; assigning ID " + (this.windowList.length - 1), "forkwindow-debug");
                this.nextWindowID++;
            }

            dumpToConsole("Load of " + requestURL + " in window " + this.windowList.indexOf(DOMWindow));


            // Notify fork window about this transition
            if ( ! this.isReplaying() ) {

                var forkOp = this.forkOpFromMainDOMWindow(DOMWindow);

                if (! forkOp) {
                    debugger;
                } else { //if ( this.lastClickedString ) {
                    forkOp.notifyStateChange(true, true, aStatus, requestURL); 
//                 } else { // this is when there is an unknown chrome click; just load into the DOM window
//                     //                    forkOp.stepCount = 0;
//                     //                    forkOp.load(requestURL);
                }
            } // notify fork window

//             // the DOMWindow wasn't found on the active windows list, but since we determined it
//             // was triggered by a user action, then add it to the list
//             if(!found) {
//                 this.activeWindows.push(DOMWindow);
//             }

            this.addSeparatorRow();
            this.addStartRow(DOMWindow, requestURL);
            //               this.addTriggerRow(this.lastClickedString);

            //               //              newTriggerIndex = this.rows - 1;

            //               // Save the form data for the clicked-on page, which may not be
           //               // the window which received the onStateChange (if the link
            //               // opens a new window, the new window receives the notification)
            //               if (this.lastClickedWindow)
            //                   formData = saveAllFormData(htmlWinFromXULWindow(this.lastClickedWindow));
            //               this.addFormRow(formData);
	    
            // If the load came from a Chrome element then force a re-sync
            dumpToConsole("onStateChange lastclickedstring is " + this.lastClickedString, "userevent-debug");
            dumpToConsole("onStateChange lastclickedwindow is " + this.lastClickedWindow, "userevent-debug");

            // write this in so the nsICookieConsent can determine the domains of top
            // level requests (documentURI is not changed until after page is loaded)
            // we keep track of the "original top level request" to deal with multiple
            // redirects before the page is loaded. See the code in the policy object to 
            // see how this works.
            var DOMWindowUnwrapped = DOMWindow.wrappedJSObject ? 
                DOMWindow.wrappedJSObject : DOMWindow;
            DOMWindowUnwrapped.topLevelRequest = requestURL;
            aRequest.QueryInterface(Components.interfaces.nsIChannel);
            DOMWindowUnwrapped.originalTopLevelRequest = aRequest.originalURI.spec;

            // If replaying, check for consistency with the replay list
            if ( this.isReplaying() ) {
                this.replayState.startedLoading(requestURL);
            }
	    
            this.lastClickedString = null;
            this.lastClickedWindow = null;
            this.addSeparatorRow();
            this.addSeparatorRow();
            dumpToConsole(win + requestURL + ": " + parseFlags(aFlag), "statechange-debug");
            notifyObservers(aRequest,"state-change","START");
            
        } else if (aFlag & STATE_STOP) {

            // This URL check is because we were getting spurious stops for sub-window 
            // content (or something)
            if( equalsURL(requestURL,DOMWindow.location.href ) &&
                (aFlag & STATE_IS_WINDOW)) {
                    if(this.isReplaying()) {
                        this.replayState.stoppedLoading(DOMWindow);
                    } else {
                        var forkOp = this.forkOpFromMainDOMWindow(DOMWindow);
                        if ( forkOp ) {
                            forkOp.notifyStateChange(true, false, aStatus, requestURL);
                        }
                    }
		   
                    this.addSeparatorRow();
                    this.addStopRow(DOMWindow, requestURL, DOMWindow.document.title);
                    this.addSeparatorRow();

                    dumpToConsole(win + requestURL + ": " + parseFlags(aFlag), "statechange-debug");
                    notifyObservers(aRequest,"state-change","STOP");

                    return 0;
                }
        } else if (aFlag & STATE_REDIRECTING) {
            dumpToConsole(win + requestURL + ": " + parseFlags(aFlag),"redirects-debug");
        }

        return 0;
    };
  
    cookielog.onLocationChange = function(aProgress, aRequest, aLocation) {};
 
    return cookielog;
}

function parseFlags(aStateFlags) {
    var tag = "";
    var nsIWebProgressListener = Components.interfaces.nsIWebProgressListener;
  
    if (aStateFlags & nsIWebProgressListener.STATE_START) {
        tag += "START";
    }
    if (aStateFlags & nsIWebProgressListener.STATE_REDIRECTING) {
        tag += "REDIR";
    }
    if (aStateFlags & nsIWebProgressListener.STATE_TRANSFERRING) {
        tag += "TRANS";
    }
    if (aStateFlags & nsIWebProgressListener.STATE_STOP) {
        tag += "STOP ";
    }
  
    if (aStateFlags & nsIWebProgressListener.STATE_IS_REQUEST) {
        tag += " | IS_REQ";
    } else {
        tag += "         ";
    }
    if (aStateFlags & nsIWebProgressListener.STATE_IS_DOCUMENT) {
        tag += " | IS_DOC";
    } else {
        tag += "         ";
    }
    if (aStateFlags & nsIWebProgressListener.STATE_IS_NETWORK) {
        tag += " | IS_NET";
    } else {
        tag += "         ";
    }
    if (aStateFlags & nsIWebProgressListener.STATE_IS_WINDOW) {
        tag += " | IS_WIN";
    }
  
    return tag;
};

function getWindowIDFromRow(rowdata) {
    var tmp;
    tmp = rowdata.match(/^\[(\d+)\]/);
    return ( tmp ? tmp[1] : null );
}

// Returns hash with keys 
// iframe, mainURI (if iframe), uri, idString, isChrome, 
// eltTagName, eltID, clickX (element-relative offset), clickY (ditto)
function parseTrigger(rowdata) {
    var nextIdx = 1;
    var parts = new Array();

    parts["iframe"] = false;
    parts["mainURI"] = null;
          
    rowdata = rowdata.replace(new RegExp("^\\[\\d+\\]\\s+"+oCookieLog.triggerPrefix), "");
    
    var coords = rowdata.match(/^<(\d+),(\d+),(\d+)>/);
    if (! coords )
        throw ("Could not parse click coordinates from trigger string");
    
    parts["button"] = coords[1];
    parts["clickX"] = coords[2];
    parts["clickY"] = coords[3];

    rowdata = rowdata.replace(/^<(\d+),(\d+),(\d+)>/, "");

    var tmp;
    if (rowdata.match(/^IFRAME/))
        {
            tmp = rowdata.match(new RegExp("^IFRAME\\{([^}]*)\\}\\{([^}]*)\\}\\{(.*)\\}(.*)"));
            parts["iframe"] = true;
            parts["mainURI"] = tmp[nextIdx++];
        } else {
            tmp = rowdata.match(new RegExp("^\\{([^}]*)\\}\\{(.*)\\}(.*)"));
        }

    parts["uri"] = tmp[nextIdx++];
    var outerHTML = tmp[nextIdx++];
    var nodeStack = tmp[nextIdx++];
    parts["idString"] = "{" + outerHTML + "}" + nodeStack;

    parts["isChrome"] = isChromeURL(parts["uri"]);


    tmp = nodeStack.match(new RegExp("^([^(]+)\\(([^)]*)\\)"));
    parts["eltTagName"] = tmp[1];
    parts["eltID"] = tmp[2];

    return parts;
};

