// -*- mode: C++; c-basic-offset: 4; fill-column: 75; tab-width: 4; indent-tabs-mode: nil -*-
//
//  **** BEGIN LICENSE BLOCK ****
//  Copyright(c) 2005 Umesh Shankar and Chris Karlof
//  This program is free software; you can redistribute it and/or modify it under
//  the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
//  more details.
//
//  You should have received a copy of the GNU General Public License along with
//  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
//  Place, Suite 330, Boston, MA 02111-1307 USA
//  **** END LICENSE BLOCK ****
//
//

if (window.content) {
    cookieOverlay.init();
} else {
    window.addEventListener("load", function() { cookieOverlay.init(); }, false);
}

var cookieOverlay = {
    blockUser: false,

    blockEvent:  function(aEvent) {
        if ( this.blockUser ) {
            aEvent.stopPropagation();
            aEvent.preventDefault();
            return false;
        }
    },

    consoleDumpStream: null,

    init: function() {
        if ( window.name.match(/^fork\d+$/) ) {
            return;
        }

        // We want to see these events in the bubbling phase
        // since we want default actions to have their effect
        // first
        addBindTo(this.onClick);
        addBindTo(this.onSubmit);
        addBindTo(this.blockEvent);

        window.addEventListener("click", this.onClick.bind(this), true);
        window.addEventListener("submit", this.onSubmit.bind(this), true);

        var Branch = getCookielogBranch();
        Branch.QueryInterface(Components.interfaces.nsIPrefBranch2);
        Branch.addObserver("debug", this, false);

        try {
            if ( Branch.getBoolPref("userstudy") ) {
                document.getElementById("debug-checkbox").style.visibility = "collapse";
                document.getElementById("paranoia-button").style.visibility = "collapse";           
                
//                 var profd = getSpecialDir("ProfD");
//                 var dateStr = ((new Date()).toLocaleString()).replace(/[\s,l\\\'\"()]+/g, "-");
//                 profd.append("javascript-console-" + dateStr);
//                 //file.create(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0664);
//                 this.consoleDumpStream = openForWriting(profd,false);                
            }
        } catch(e) {
            dumpToConsole("This profile does not have a cookielog.userstudy preference for conducting user studies");
            //            dumpToConsole("Exception text: " + e.toString());
        }
        
        if ( ! Branch.prefHasUserValue("debug") ) {
            Branch.setBoolPref("debug", false);
        }
        var cb = window.document.getElementById("debug-checkbox");
        cb.checked = Branch.getBoolPref("debug");
       

        addObserver(this, "compare-dialog-on", false);
        addObserver(this, "compare-dialog-off", false);
        addObserver(this, "set-statusbar-sync", false);

        window.addEventListener("mousedown", this.blockEvent.bind(this), true);
        window.addEventListener("mouseup", this.blockEvent.bind(this), true);
        window.addEventListener("keydown", this.blockEvent.bind(this), true);
        window.addEventListener("keyup", this.blockEvent.bind(this), true);
        window.addEventListener("focus", this.blockEvent.bind(this), true);
        window.addEventListener("blur", this.blockEvent.bind(this), true);
        window.addEventListener("click", this.blockEvent.bind(this), true);


        // Replace the BrowserBack() function, and other navigation functions so we can capture the event
        // no matter how it's invoked. Ideally we'd just replace the Browser:Back command, but the toolbar
        // does not use the <command>; it just calls BrowserBack() directly.

        var realBrowserBack = BrowserBack;
        BrowserBack = function(aEvent, aIgnoreAlt) {
            notifyObservers(htmlWinFromXULWindow(window), "trigger-domwindow", null);
            notifyObservers(window, "trigger", "<0,0,0>{chrome://browser/content/browser.xul}{tagName=xul%253Atoolbarbutton&label=Back}xul:toolbarbutton()");
            if ( whereToOpenLink(aEvent, false) != "current" ) {
                // We don't support opening links in new windows from Back button
                return;
            }
            realBrowserBack(aEvent, aIgnoreAlt);
        }

        // Now do the same for Forward and Reload
        var realBrowserForward = BrowserForward;
        BrowserForward = function(aEvent, aIgnoreAlt) {
            notifyObservers(htmlWinFromXULWindow(window), "trigger-domwindow", null);
            notifyObservers(window, "trigger", "<0,0,0>{chrome://browser/content/browser.xul}{tagName=xul%253Atoolbarbutton&label=Forward}xul:toolbarbutton()");
            if ( whereToOpenLink(aEvent, false) != "current" ) {
                // We don't support opening links in new windows from Forward button
                return;
            }
            realBrowserForward(aEvent, aIgnoreAlt);
        }


        var realBrowserReload = BrowserReload;
        BrowserReload = function(aEvent, aIgnoreAlt) {
            notifyObservers(htmlWinFromXULWindow(window), "trigger-domwindow", null);
            notifyObservers(window, "trigger", "<0,0,0>{chrome://browser/content/browser.xul}{tagName=xul%253Atoolbarbutton&label=Reload}xul:toolbarbutton()");
            if ( whereToOpenLink(aEvent, false) != "current" ) {
                // We don't support opening links in new windows from Reload button
                return;
            }
            realBrowserReload(aEvent, aIgnoreAlt);
        }


        // Open the extension window
        var ww = getWindowWatcher();
        if ( ! ww.getWindowByName("cookielog", null) ) {
            window.setTimeout (function() {
                window.openDialog("chrome://doppelganger/content/LiveHTTPHeaders.xul", "cookielog", "chrome,dialog=no,width=0,height=0,resizable");
            }, 2000);
        }

    },

    observe: function(aSubject, aTopic, aData) {
        var NS_PREFBRANCH_PREFCHANGE_TOPIC_ID = "nsPref:changed";

        if ( aTopic == "compare-dialog-on" ) { // block keys user actions comparison dialog is up
            this.blockUser = true;
        } else if ( aTopic == "compare-dialog-off" ) {
            this.blockUser = false;
        } else if ( aTopic == NS_PREFBRANCH_PREFCHANGE_TOPIC_ID ) {
            var cb = document.getElementById('debug-checkbox');
            cb.checked = getCookielogBranch().getBoolPref("debug");
        } else if ( aTopic == "set-statusbar-sync" ) {
            if ( htmlWinFromXULWindow(window) != aSubject ) {
                return;
            }

            var elt = document.getElementById("doppelStatus");           
            var fixme = document.getElementById("fixme-button");
            fixme.disabled = false;
            fixme.tooltipText = "";
            switch ( aData  ) {
                case "SYNC" : 
                    elt.value = "Synced";
                    elt.style.color = "green";
                    break;

                case "NOSYNC" : 
                    elt.value = "No Sync";
                    elt.style.color = "red";
                    break;

                case "LOADING" : 
                    elt.value = "Loading...";
                    elt.style.color = "yellow";
                    fixme.disabled = true;
                    fixme.tooltipText = "FIXME is disabled while the fork window is still loading (not synced yet)";
                    break;

                case "NOFORK" : 
                    elt.value = "Fork off";
                    elt.style.color = "#BBBBBB";

                    break;            
            }
        }
    },

    fixMe: function() {
        notifyObservers(window, "fixme", "");
    },

    syncToFork: function() {
        var url = htmlWinFromXULWindow(window).location;
        notifyObservers(window, "sync-to-fork", url.href);
    },
  
    onSubmit: function(aEvent) {
        var frm = aEvent.originalTarget;
        if ( frm.tagName.toUpperCase() != "FORM" ) {
            frm = frm.form;
        }

        if ( frm.tagName.toUpperCase() != "FORM" ) {
            dumpToConsole("Internal Error: onSubmit called on non-form element");
            return;
        }

        for ( var i = 0; i < frm.elements.length; ++i ) {
            if ( frm.elements[i].type && frm.elements[i].type.toLowerCase() == "password" ) {
                if ( frm && frm.action ) {
                    var tailDomain;
                    if ( frm.action.match(/^https?:/i) ) {
                        tailDomain = getTailURLDomain(getURLDomain(frm.action));
                    } else {
                        tailDomain = getTailURLDomain(frm.ownerDocument.defaultView.location.hostname);
                    }
                    notifyObservers(frm, "login-form-detected", tailDomain);
                }                
            }
        }
    },

    //
    // We trap clicks so that we can remember which element caused the
    // next page load (for replaying later), and to distinguish which
    // pages were loaded due to a user event.
    //

    onClick: function(aEvent) {
        if ( this.blockUser ) {
            aEvent.stopPropagation();
            aEvent.preventDefault();
            return false;
        }

        // We only handle left clicks; other stuff is too unpredictable
        if ( aEvent.button != 0 ) {
            return true;
        }

        var wasChrome = false;
	    
        if ( isChromeURL(aEvent.originalTarget.baseURI) ) {
            wasChrome = true;
        }
	
        var cookieLogLastClicked = aEvent.originalTarget;           
        if (! cookieLogLastClicked.ownerDocument) {
            return true;
        }

        var lastClickedWindow = window;
        var lastClickedDOMWindow;
        var lastClickedURI;
        var iframeStr = "";
        var locationStr = "";
	
        if (! wasChrome ) {
            var lastClickedDOMWindow = aEvent.originalTarget.ownerDocument.defaultView;
            var lastClickedURI = lastClickedDOMWindow.location;
            // Find the element which actually did something when
            // clicked: it should have an onclick or an href or
            // be an INPUT element
            var objtocheck = cookieLogLastClicked.wrappedJSObject ? 
                cookieLogLastClicked.wrappedJSObject : cookieLogLastClicked;

            while (! wasChrome &&
                   ! objtocheck.onclick &&
                   ! objtocheck.onfocus &&
                   ! objtocheck.onchange && 
                   ! objtocheck.href &&
                   ! objtocheck.form &&
                   objtocheck.tagName &&
                   (objtocheck.tagName != "INPUT" && objtocheck.tagName != "SELECT") )
                {
                    cookieLogLastClicked = cookieLogLastClicked.parentNode;
                    if (! cookieLogLastClicked) 
                        return true;

                    objtocheck = cookieLogLastClicked.wrappedJSObject ? 
                        cookieLogLastClicked.wrappedJSObject : cookieLogLastClicked;
                }


            // Figure out if the click was in a frame, and record
            // the main page's URL as well if so
            var mainURI = lastClickedDOMWindow.top.location;
	
            if (mainURI != lastClickedURI) {
                iframeStr = "IFRAME{"+mainURI.toString()+"}";
            }

            // Get the click's document-relative offset
            var docX = aEvent.clientX + lastClickedDOMWindow.pageXOffset;
            var docY = aEvent.clientY + lastClickedDOMWindow.pageYOffset;
            //dumpToConsole("clickclientx = " + aEvent.clientX + "; clickclienty = " + aEvent.clientY);
            //dumpToConsole("clickdocx = " + docX + "; clickdocy = " + docY);
	    
            // ...and the element's doc-relative offset
            var eltDocOffset = getDocOffset(aEvent.originalTarget);
            //dumpToConsole("eltx = " + aEvent.originalTarget.offsetLeft + "; elty = " + aEvent.originalTarget.offsetTop);
            //dumpToConsole("eltdocx = " + eltDocOffset["x"] + "; eltdocy = " + eltDocOffset["y"]);

            // For certain cases, like SELECT elements that pop down a
            // list, these values can be negative. I think that's
            // because the dropdown counts as a separate window or
            // something for the purposes of coordinates. In any
            // case, we'll just cap them to 0.
            var eltRelX = (docX - eltDocOffset["x"]);
            var eltRelY = (docY - eltDocOffset["y"]);

            // Weird case: For SELECT elements, *non*-negative coordinates
            // cause the box to drop down -- but we don't want to transmit
            // those clicks because they change focus and kill the dropdown
            if ( objtocheck.tagName == "SELECT" && objtocheck.type.match(/^select-one/i) &&
                 objtocheck.size <= 1 ) {
                if (eltRelX >= 0 && eltRelY >= 0)
                    return true;
            }

            if (eltRelX < 0 || eltRelY < 0) {
                eltRelX = 0;
                eltRelY = 0;
            }

            // Clicking on scrollbars in SELECT dropdowns causes NaN coordinates
            if (! isFinite(eltRelX) || ! isFinite(eltRelY) )
                return true;
	    
            // click coords - element's coords = element-relative coords
            locationStr = "<" + aEvent.button + "," + eltRelX + "," + eltRelY + ">";
        } else { // Chrome
            // For chrome, just stick with (0,0)
//             var lastClickedDOMWindow = htmlWinFromXULWindow(window);
//             var lastClickedURI = window.location;
//             locationStr = "<" + aEvent.button + ",0,0>";

            return true;
        }

        //                 dumpToConsole("clicked on a " + aEvent.originalTarget.tagName );
               
        if ( cookieLogLastClicked ) {            
            dumpToConsole (locationStr + iframeStr + "{" + lastClickedURI.toString() + "}" + nodeIDString(cookieLogLastClicked), "userevent-debug");
            notifyObservers(lastClickedDOMWindow, "trigger-domwindow", null);
            notifyObservers(window, "trigger", locationStr + iframeStr + "{" + lastClickedURI.toString() + "}" + nodeIDString(cookieLogLastClicked));
        }

        return true;
    }
}

// addEventListener("load", gMyBrowserProgressListener.onInit, false);
// addEventListener("unload", gMyBrowserProgressListener.onUninit, false);

// var gMyBrowserProgressListener = {
//     onInit: function() {
// 	mydocloader.addProgressListener(gMyBrowserProgressListener,Components.interfaces.nsIWebProgress.NOTIFY_ALL);
//     },
	
//     onUninit:  function() {
// 	mydocloader.removeProgressListener(gMyBrowserProgressListener);
//     },

//     mydocloader: Components.classes["@mozilla.org/docloaderservice;1"].getService(Components.interfaces.nsIWebProgress),

//     activeWindows: new Array(),

//     stateIsRequest: false,
//     //  loadGroup:null,
  
//     QueryInterface : function(aIID) {
// 	if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
// 	    aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
// 	    aIID.equals(Components.interfaces.nsISupports))
// 	return this;
// 	throw Components.results.NS_NOINTERFACE;
//     },
  
//     onProgressChange : function(aProgress, aRequest, aCurSelfProgress, aMaxSelfProgress, aCurTotalProgress, aMaxTotalProgress) {
//     },
  
//     onStatusChange : function(aProgress, aRequest, aStatus, aMessage) {
//     },
  
//     onSecurityChange : function(a,b,c){},
//     onLinkIconAvailable : function(a){},

//     onStateChange: function(aProgress, aRequest, aFlag, aStatus) {
//     },
    
//     onLocationChange : function(aProgress, aRequest, aLocation) {
//     },



//     parseFlags: function(aStateFlags) {
// 	var tag = "";
// 	var nsIWebProgressListener = Components.interfaces.nsIWebProgressListener;
  
// 	if (aStateFlags & nsIWebProgressListener.STATE_START) {
// 	    tag += "START";
// 	}
// 	if (aStateFlags & nsIWebProgressListener.STATE_REDIRECTING) {
// 	    tag += "REDIR";
// 	}
// 	if (aStateFlags & nsIWebProgressListener.STATE_TRANSFERRING) {
// 	    tag += "TRANS";
// 	}
// 	if (aStateFlags & nsIWebProgressListener.STATE_STOP) {
// 	    tag += "STOP ";
// 	}
  
// 	if (aStateFlags & nsIWebProgressListener.STATE_IS_REQUEST) {
// 	    tag += " | IS_REQ";
// 	} else {
// 	    tag += "         ";
// 	}
// 	if (aStateFlags & nsIWebProgressListener.STATE_IS_DOCUMENT) {
// 	    tag += " | IS_DOC";
// 	} else {
// 	    tag += "         ";
// 	}
// 	if (aStateFlags & nsIWebProgressListener.STATE_IS_NETWORK) {
// 	    tag += " | IS_NET";
// 	} else {
// 	    tag += "         ";
// 	}
// 	if (aStateFlags & nsIWebProgressListener.STATE_IS_WINDOW) {
// 	    tag += " | IS_WIN";
// 	}
  
// 	return tag;
//     }
// }
