// -*- mode: C++; c-basic-offset: 4; fill-column: 75; tab-width: 4; indent-tabs-mode: nil -*-
//
//  **** BEGIN LICENSE BLOCK ****
//  Copyright(c) 2005 Umesh Shankar and Chris Karlof
//  This program is free software; you can redistribute it and/or modify it under
//  the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
//  more details.
//
//  You should have received a copy of the GNU General Public License along with
//  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
//  Place, Suite 330, Boston, MA 02111-1307 USA
//  **** END LICENSE BLOCK ****
//


const COOKIECONSENT_CONTRACTID = "@mozilla.org/cookie-consent;1";
const COOKIECONSENT_DESCRIPTION = "Cookie consent service";

const COOKIE_STATUS_ACCEPTED = 1;
const COOKIE_STATUS_DOWNGRADED = 2;
const COOKIE_STATUS_REJECTED = 4;

function MakeCookiePermissions() {
    // HACK we shouldn't have to do this just to make a table
    var cookieperms = new HeaderInfoLive();

    cookieperms.domains = new Array();
    cookieperms.permissions = new Array();

    cookieperms.enableFPSession = false;
    cookieperms.enableFPPersistent = false;
    cookieperms.enableTPSession = false;
    cookieperms.enableTPPersistent = false;

    cookieperms.toggleFPSession  = function(state) { this.enableFPSession = state; };
    cookieperms.toggleFPPersistent  = function(state) { this.enableFPPersistent = state; };
    cookieperms.toggleTPSession  = function(state) { this.enableTPSession = state; };
    cookieperms.toggleTPPersistent  = function(state) { this.enableTPPersistent = state; };

    cookieperms.cookiePerms = new Array();

    cookieperms.observe = function(aSubject, aTopic, aData) {
        if (aTopic == "cookie-getter") {
            aSubject.QueryInterface(Components.interfaces.nsIMutableArray);
            aSubject.enumerate();
            var channel = aSubject.queryElementAt(0, Components.interfaces.nsIChannel);
            var cookieString = aData;
            
            var domWindow = getDOMWindowFromChannel(channel);

            if ( ! domWindow) {
                return;
                //                alert("NULL domWindow in cookie-getter");
                //                debugger;
            }

            var forkOp = domWindow && oCookieLog.forkOpFromDOMWindow(domWindow.top);

            var prefixString = "FORK"; // + forkOp.ID;
            var prefixRE = new RegExp("^" + prefixString);
            var returnCookieString = "";
            var cookies = cookieString.split(/;\s*/);
            var i;
            var modified = false;

            for (i in cookies) {
                
                if ( prefixRE.test(cookies[i]) ) { // special FORK cookie

                    if ( forkOp ) {
                        if ( returnCookieString.length != 0 ) {
                            returnCookieString += "; ";
                        }
                        
                        returnCookieString += cookies[i].substr(prefixString.length);
                    }
                    
                    modified = true;
                    
                } else { // ordinary cookie

                    if ( ! forkOp ) {
                        if ( returnCookieString.length != 0 ) {
                            returnCookieString += "; ";
                        }
                        returnCookieString += cookies[i];
                    } else {
                        modified = true;
                    }                           
                }

            }

            
            if ( domWindow == domWindow.top ) {
                var finalString = (modified ? returnCookieString : cookieString);
                var finalCookies = finalString.split(';');
                dumpToConsole("Cookies being sent for " + (forkOp ? "fork" : "main") + " window: " + 
                              finalCookies.join("\n"), "cookie-debug" );
            }

            if ( modified ) {
                var returnCookieStringVar = Components.classes["@mozilla.org/variant;1"]
                .createInstance(Components.interfaces.nsIWritableVariant);
                returnCookieStringVar.setAsString(returnCookieString);
                aSubject.appendElement(returnCookieStringVar, false);
            }
        }
    };

    cookieperms.start = function() {
        this.oDump = document.getElementById("cookiepermissions");
        this.oDump.treeBoxObject.view = this;

        this.hScrollBar = document.getElementById("cookiepermissions-scroll");
        setInterval(this.hScrollHandler,100);

        this.initAtoms();

        this.loadPermissions();

        addObserver(this, "cookie-getter", false);
    };

    cookieperms.getCellText = function(row, column) {

        if(column.id == "cookiepermissions-domain") {
            return this.domains[row];
        } else if (column.id == "cookiepermissions-permissions") {
            return this.permissions[row];
        } else {
            dumpToConsole("Error in get CellTest: " + column.id);
            return null;
        }
    };

    cookieperms.addRow = function(domain, permissions, type) { 
        var style;

        // remove a row with no permissions
        for( var i=0; i < this.rows; ++i) {
            if(domain == this.domains[i]) {
                if(permissions === "") {
                    this.rows--;
                    this.domains.splice(i,1);
                    this.permissions.splice(i,1);
                } else {
                    this.permissions[i] = permissions;
                }
                return;
            }
        }

        // don't add a row with no permissions
        if(permissions === "") {
            return;
        }

        // Compute the style
        if (this.rows>0) {
            var ptype = this.type[this.rows-1];
            var pstyle = this.style[this.rows-1];
            if (type != ptype) {
                // We have a new type.  The line above may be single or last.
                if (pstyle == this.FIRST) {
                    pstyle = this.SINGLE;
                } else {
                    pstyle = this.LAST;
                }
                this.style[this.rows-1] = pstyle; // Update the style for the line above
                style = this.FIRST;
            } else {
                // Same type.  This line should be middle
                style = this.MID;
            }
        } else {
            // The first line.  No doubts, it should be of style FIRST.
            style = this.FIRST;
            document.getElementById(this.datapresent).removeAttribute('disabled');
        }

        // Add the row
        this.type.push(type);
        this.style.push(style);
        this.rows = this.domains.push(domain);
        this.permissions.push(permissions);
    };

    cookieperms.onselect = function() {
        var selection = this.oDump.view.selection;
        var index = selection.currentIndex;
        document.getElementById("cookiepermissions-domain-input").value = this.domains[index];
    };

    cookieperms.savePermissions = function() {
        var profd = getSpecialDir("ProfD");
        profd.append("cookielog");
        //file.create(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0664);
        var foStream = openForWriting(profd,false);
        var data = "# CookieLog Domain Preferences\n";
        for(var i=0;i<this.rows;i++) {
            data += this.domains[i] + " " + this.permissions[i] + "\n";
        }
        foStream.write(data, data.length);
        foStream.close();
    };

    cookieperms.loadPermissions = function() {
        var profd = getSpecialDir("ProfD");
        profd.append("cookielog");
        if(!profd.exists()) {
            return;
        }
        // open an input stream from file
        var istream = Components.classes["@mozilla.org/network/file-input-stream;1"].
        createInstance(Components.interfaces.nsIFileInputStream);
        istream.init(profd, 0x01, 0444, 0);
        istream.QueryInterface(Components.interfaces.nsILineInputStream);

        // read lines into array
        var line = {}, lines = [], hasmore;
        do {
            hasmore = istream.readLine(line);
            lines.push(line.value); 
        } while(hasmore);
    
        istream.close();   

        // first line of preferences file is a comment
        if(lines.length > 1) {
            for(var i=1;i<lines.length;i++) {
                var tmp = lines[i].match(/(\S*)\s(.*)/);
                //dumpToConsole("LOADED " + tmp[1] + ": " + tmp[2]);
                this.addRow(tmp[1],tmp[2],oCookieLog.COOKIE);
                this.restoreCookiePerms(tmp[1],tmp[2]);
                //      dumpToConsole("LOADED " + tmp[1] + ": " + tmp[2]);
            }
        }
        if(lines) {
            this.rowCountChanged(0,lines.length-1);
        }
    };

    cookieperms.restoreCookiePerms = function(domain, strCookiePermissions) {
        var perms = parseCookiePermissionsToInt(strCookiePermissions);
        this.cookiePerms[domain] = perms;
        var uri = createURI();
        uri.spec = domain;
        //    var permMan = getPermissionManager();
        //permMan.add(uri,"cookie",permMan.ALLOW_ACTION);
    };

    cookieperms.setDomainCookiePermsWrapper = function() {
        var domain = document.getElementById("cookiepermissions-domain-input").value;
        var domainPerms = 0;
        if(this.enableFPSession) {
            domainPerms |= FP_SESSION;
        }
        if(this.enableFPPersistent) {
            domainPerms |= FP_PERSISTENT;
        }
        if(this.enableTPSession) {
            domainPerms |= TP_SESSION;
        }
        if(this.enableTPPersistent) {
            domainPerms |= TP_PERSISTENT;
        }
        this.setDomainCookiePerms(domain,domainPerms);
    }

    cookieperms.setForkCookiePermsWrapper = function() {
        var domainPerms = 0;
        if(this.enableFPSession) {
            domainPerms |= FP_SESSION;
        }
        if(this.enableFPPersistent) {
            domainPerms |= FP_PERSISTENT;
        }
        if(this.enableTPSession) {
            domainPerms |= TP_SESSION;
        }
        if(this.enableTPPersistent) {
            domainPerms |= TP_PERSISTENT;
        }
        this.setForkCookiePermissions(oCookieLog.forkOperations.getDOMWindow(),domainPerms);
        if(domainPerms == 0) {
            document.getElementById("fork-window-permissions").value = "";
        } else {
            document.getElementById("fork-window-permissions").value = parseCookiePermissionsToString(domainPerms);
        }
    }
    
    cookieperms.setDomainCookiePerms = function (domain, intDomainPerms) {
        // FIXME: handling cookie permissions in the browser in the way will not enable us to log cookies
        // set by javascript, since they will be denied. this may be a weird corner case though
        if(domain && intDomainPerms != null) {
            this.cookiePerms[domain] = intDomainPerms;
            // dont do this anymore --> replaced by using nsICookieConsent
            //  var uri = createURI();
            //  uri.spec = domain;
            //    var permMan = getPermissionManager();
            //      if(domainPerms & FP_SESSION || domainPerms & FP_PERSISTENT) {
            // 	permMan.add(uri,"cookie",permMan.ALLOW_ACTION);
            //       } else if(permMan.testPermission(uri,"cookie") == permMan.ALLOW_ACTION) {
            // 	permMan.remove(domain,"cookie");
            //       }
            var oCookiePermissionsOldRows = this.rowCount;
            this.addRow(domain,parseCookiePermissionsToString(intDomainPerms),oCookieLog.COOKIE);
            this.rowCountChanged(oCookiePermissionsOldRows,(this.rows-oCookiePermissionsOldRows));
        } else {
//             alert("INTERNAL ERROR: setDomainCookiePerms(); null or undefined domain or permission set: " + domain);
//             dumpToConsole("INTERNAL ERROR: setDomainCookiePerms(); null or undefined domain or permission set: " + domain);
//             debugger;
        }
        //    alert(domain + ": " + this.cookiePerms[domain]);
    };

    cookieperms.getDomainCookiePerms = function (domain) {
        if(this.cookiePerms[domain]) {
            return this.cookiePerms[domain];            
        } else {
            return 0;
        }        
    };

//     // Like setDomainCookiePerms, but just adds the specified permissions to those already set
//     cookieperms.addDomainCookiePerms = function (domain, intDomainPerms) {
//         this.setDomainCookiePerms(domain, this.cookiePerms[domain] | intDomainPerms);
//     };

    cookieperms.setForkCookiePermissions = function(forkWindow, intCookiePermissions) {
        var forkOp = oCookieLog.forkOpFromDOMWindow(forkWindow);

        if (! forkOp) {
            alert("Problem: no forkOperations object for a fork window");
        }

        forkOp.cookiePermissions = intCookiePermissions;
    };

    cookieperms.getPermissions = function(DOMWindow,documentDomain) {
        // try to find permissions for this DOMWindow in the
        // forked window list
        var forkOp = oCookieLog.forkOpFromDOMWindow(DOMWindow);

        if (forkOp) {
            return forkOp.cookiePermissions;
        }

        // otherwise just return the general permissions for this domain
        if(this.cookiePerms[documentDomain]) {
            return this.cookiePerms[documentDomain];
        } else {
            return 0;
        }
    };

    cookieperms.clearAllPerms = function() {
        this.cookiePerms = new Array;
        this.permissions = new Array;
        this.clear();
        this.savePermissions();
    }
    
    return cookieperms;
}

var policy = {
    getConsent: function(uri, oHttp, isForeign, arg_cookieString, policy) {

        // Shortcut-accept our explicitly added cookies (identified due to
        // lack of a channel)
        if (! oHttp) {
            return COOKIE_STATUS_ACCEPTED;
        }
        
        if ( ! uri.spec.match(/^http/) ) {
            return COOKIE_STATUS_ACCEPTED;
        }

        var name = canonicalizeURL(uri.spec);
        var documentName = null;
        var documentNameSrc = null;
        var documentDomain = null;
        var found = false;
        var topDOMWindow = null; 
        var DOMWindow = null;
        var result = null;
        var progress = null;
        var forkOp = null; // if a fork window, its ForkOperations object

        // FIXME: For some reason even reading cookieString directly seems to cause problems
        // Perhaps I am not passing it in properly
        var cookieString = (arg_cookieString == null) ? null : arg_cookieString + "";

        if (cookieString) {
            dumpToConsole("Cookie string is <" + cookieString + ">", "cookie-debug");
        } else {
            dumpToConsole("URL for sending cookies in getConsent is " + uri.spec, "cookie-debug");
        }

        
        // Shortcut-accept the special FORK cookies
        // FIXME: this doesn't check that it isn't coincidentally named that way
        if (cookieString) {
            var re = new RegExp("^FORK");
            var id = re.exec(cookieString);
            if (id) {
                return COOKIE_STATUS_ACCEPTED;
            }
        }

        // We transform accepted fork-window cookies into differently-named cookies
        // so we can keep track of them
        var returnDecision = function(decision) {
            if (decision != COOKIE_STATUS_REJECTED && forkOp && cookieString) {
                // FIXME: This approach can push a cookie over the 4K limit
                // FIXME: This approach allows tampering from websites
                var cookies = cookieString.split(/\n/);
                for (var i in cookies) {                    
                    getCookieService().setCookieString(uri, null, 
                                                       "FORK" + cookies[i], 
                                                       // "FORK" + forkOp.ID + cookies[i], 
                                                       oHttp);                    
                }

                return COOKIE_STATUS_REJECTED;
            }

            return decision;
        };
	
        oHttp.QueryInterface(Components.interfaces.nsIHttpChannelInternal);

        progress = getWebProgressFromChannel(oHttp);

        // determine the DOMWindow for this request
        if(progress && progress.DOMWindow) {
            DOMWindow = progress.DOMWindow;
            topDOMWindow = DOMWindow.top;
//             for(i=0;i<oCookieLog.activeWindows.length;i++) {
//                 if(oCookieLog.activeWindows[i] == DOMWindow.top) {
//                     topDOMWindow = DOMWindow.top;
//                     found = true;
//                     break;
//                 }
//             }
//             if(!found) {
//                 for(i=0;i<oCookieLog.activeWindows.length;i++) {
//                     if(DOMWindow.opener && (oCookieLog.activeWindows[i] == DOMWindow.opener.top)) {
//                         topDOMWindow = DOMWindow.opener.top;
//                         found = true;
//                         break;
//                     }
//                 }	
//             }

            // check for DOMWindow in list of fork windows
            for(i in oCookieLog.forkOperationsList) {
                if(topDOMWindow == oCookieLog.forkOperationsList[i].getDOMWindow()) {
                    forkOp = oCookieLog.forkOperationsList[i];
                    break;
                }
            }

        } else {
            dumpToConsole("No DOMWindow in WebProgress for: " + name,"domwindow-debug");
        }    


        // if we couldn't find a DOMWindow, then try alternative methods to 
        // determine the document URI for this request
        if(!topDOMWindow) {
            if(! (name.indexOf("?") < 0 && name.match(/favicon\.ico|nav_.*\.gif/))) { 
                dumpToConsole("OBSERVE RESPONSE: Cannot find top DOMWindow for " + name,"domwindow-debug");
            } else {
                // ignore requests for these things
                return returnDecision(COOKIE_STATUS_REJECTED);
            }
	    
            if(oHttp.documentURI) {
                documentName =  oHttp.documentURI.spec;
                documentNameSrc = "DOCUMENT_URI: ";
            } else if (oHttp.loadGroup) {
                if(oHttp.loadGroup.defaultLoadRequest) {
                    if(oHttp.loadGroup.defaultLoadRequest.documentURI) {
                        documentName = oHttp.loadGroup.defaultLoadRequest.documentURI.spec;
                        documentNameSrc = "DEFAULT_LOAD_REQUEST_DOCUMENT_URI: ";
                    } else {
                        try { 
                            documentName = oHttp.loadGroup.defaultLoadRequest.name;
                            documentNameSrc = "DEFAULT_LOAD_REQUEST_NAME: ";
                        }
                        catch (e) { // this is a weird case, like when doing View Source
                            returnDecision(COOKIE_STATUS_ACCEPTED);
                        }
                    }
                }
            }
            dumpToConsole("Trying alternative: " + documentNameSrc + documentName,"domwindow-debug");
            if(!documentName || documentName.indexOf("http") != 0) {
                if(! (name.indexOf("?") < 0 && name.match(/favicon\.ico|nav_.*\.gif/))) { 
                    dumpToConsole("ERROR: Could not determine any DOCUMENT_URI for " + name,"domwindow-debug");
                }
                return returnDecision(COOKIE_STATUS_REJECTED);
            }
            if(!documentName) {
                dumpToConsole("null document name (not DOMWINDOW) for " + name, "domwindow-debug");
                return returnDecision(COOKIE_STATUS_REJECTED);
            }
        } else { // get documentURI from DOMWindow
            var topDOMWindowUnwrapped = topDOMWindow.wrappedJSObject ? 
            topDOMWindow.wrappedJSObject : topDOMWindow;
            // deal with top level requests and redirects of toplevel requests
            if(topDOMWindowUnwrapped.topLevelRequest) {
                var origname = canonicalizeURL(oHttp.originalURI.spec);
                // this clause seems weird, but the second clause is needed to when an unload handler is called on a page and
                // makes a request. e.g., yahoo makes an unload request. suppose we request amazon. the topLevelRequest is set
                // to amazon, but the request for the yahoo web bug comes before the request for the amazon home page. If the
                // web bug is redirected, then we don't want to be fooled into thinking that we are being redirected for amazon.

                if(origname &&
                   equalsURL(origname,topDOMWindowUnwrapped.originalTopLevelRequest) && 
                   ! equalsURL(name,origname)) {
                    dumpToConsole("Redirect " + (forkOp ? " [FORK] " : "[MAIN]") + ": changing\n" + origname + " to\n" + name,"redirects-debug");

                    // Send the relevant fork operations object a START event due to the redirect
                    if ( topDOMWindowUnwrapped.topLevelRequest != name ) {

                        if ( forkOp ) {
                            dumpToConsole("Issuing START event to Fork window for FORK because of redirect in getConsent()", "statechange-debug");
                            forkOp.notifyStateChange( false, true, NS_OK, name, null, true );
                        } else {
                            var relatedFork = oCookieLog.forkOpFromMainDOMWindow(topDOMWindow);
                            if ( relatedFork ) {
                                dumpToConsole("Issuing START event to Fork window for MAIN because of redirect in getConsent()", "statechange-debug");
                                relatedFork.notifyStateChange ( true, true, NS_OK, name, null, true );
                            } else {
                                dumpToConsole("COULD NOT issue START event to Fork window for MAIN because of redirect in getConsent(): no fork window found", "statechange-debug");
                            }
                        }

                    }

                    topDOMWindowUnwrapped.topLevelRequest = name;
                }		
                documentName = topDOMWindowUnwrapped.topLevelRequest;
            } else {
                documentName = topDOMWindow.location.href;
            } 

            if(!documentName) {
                dumpToConsole("null topLevelRequest (DOMWINDOW) for " + name, "domwindow-debug");
                return returnDecision(COOKIE_STATUS_REJECTED);
            }
        }

        // ignore this
        if(documentName.match(/^about/)) {
            return returnDecision(COOKIE_STATUS_REJECTED);
        }

        documentName = canonicalizeURL(documentName);
        documentDomain = getTailURLDomain(getURLDomain(documentName));
        var urlDomain = getURLDomain(name);

        if ( urlDomain.indexOf(documentDomain) != -1 ) {
            var perms = oCookiePermissions.getPermissions(topDOMWindow,documentDomain);
            dumpToConsole(name + "\nDocumentDomain: " + documentDomain,"domains-debug");

            if ( forkOp ) {
                return returnDecision(COOKIE_STATUS_ACCEPTED);
            }

            if ( perms & FP_PERSISTENT ) { 
                dumpToConsole("FP ACCEPTED " + "(" + (topDOMWindow ? topDOMWindow.name : "") + ") " + name,"cookie-debug");
                return returnDecision(COOKIE_STATUS_ACCEPTED);
            } else if ( perms & FP_SESSION ) {
                dumpToConsole("FP ACCEPTED (but only for later forking) "  + "(" + (topDOMWindow ? topDOMWindow.name : "") + ") " + name,"cookie-debug");
                return returnDecision(COOKIE_STATUS_ACCEPTED);
            } else {
                dumpToConsole("FP REJECTED " + "(" + (topDOMWindow ? topDOMWindow.name : "") + ") " + name,"cookie-debug");
                return returnDecision(COOKIE_STATUS_REJECTED);
            }
        } else if ( !isForeign && (topDOMWindow == DOMWindow) && oHttp.documentURI ) {
            // this case is for when getConsent is called before the DOMWindow is determined
            // this usually happens on a new request, and the browser needs to determine whether
            // to send cookies for a top level request, but the call to getConsent is done before
            // the start event. the above condition checks to see if the browser thinks it's a first
            // party cookie, but makes sure its not in an iframe
            // this oHttp.documentURI check is needed because in the mozilla src, isForeign is set incorrectly if
            // it is null
            var tailURLDomain = getTailURLDomain(urlDomain);
            var perms = oCookiePermissions.getPermissions(topDOMWindow,tailURLDomain);
            dumpToConsole(name + "\nDocumentDomain (using isForeign check): " + tailURLDomain,"domains-debug");

            if ( forkOp ) {
                return returnDecision(COOKIE_STATUS_ACCEPTED);
            }

            if( perms & FP_PERSISTENT ) {
                dumpToConsole("FP ACCEPTED (Using isForeign check) " + "(" + (topDOMWindow ? topDOMWindow.name : "") + ") " + name,"cookie-debug");
                return returnDecision(COOKIE_STATUS_ACCEPTED);
            } else if ( perms & FP_SESSION ) {
                dumpToConsole("FP ACCEPTED (but only for later forking) (Using isForeign check) " + "(" + (topDOMWindow ? topDOMWindow.name : "") + ") " + name,"cookie-debug");
                return returnDecision(COOKIE_STATUS_ACCEPTED);
            } else {
                dumpToConsole("FP REJECTED (Using isForeign check) " + "(" + (topDOMWindow ? topDOMWindow.name : "") + ") " + name,"cookie-debug");
                return returnDecision(COOKIE_STATUS_REJECTED);
            }
        } else {
            var perms = oCookiePermissions.getPermissions(topDOMWindow,documentDomain);
            dumpToConsole(name + "\nDocumentDomain: " + documentDomain,"domains-debug");
            if( (perms & TP_PERSISTENT) || (forkOp && (perms & TP_SESSION)) ) { // Fork windows need to keep even persistent cookies for later forking
                dumpToConsole("TP ACCEPTED " + "(" + (topDOMWindow ? topDOMWindow.name : "") + ") " + name,"cookie-debug");
                return returnDecision(COOKIE_STATUS_ACCEPTED);
            } else if(perms & TP_SESSION) {
                dumpToConsole("TP DOWNGRADED "  + "(" + (topDOMWindow ? topDOMWindow.name : "") + ") " + name,"cookie-debug");
                return returnDecision(COOKIE_STATUS_DOWNGRADED);
            } else {
                dumpToConsole("TP REJECTED " + "(" + (topDOMWindow ? topDOMWindow.name : "") + ") " + name,"cookie-debug");
                return returnDecision(COOKIE_STATUS_REJECTED);
            }
        } 
	
    },

    observe: function(subject, topic, data) {
	
    },

    QueryInterface: function(iid) {
        if (!iid.equals(Components.interfaces.nsISupports) &&
            !iid.equals(Components.interfaces.nsICookieConsent) &&
            !iid.equals(Components.interfaces.nsIObserver)) {
                throw Components.results.NS_ERROR_NO_INTERFACE;
            }
        return this;
    }
    
} // MakeCookieLog

var factory = {
    // nsIFactory interface implementation
    createInstance: function(outer, iid) {
        if (outer != null) throw Components.results.NS_ERROR_NO_AGGREGATION;
        return policy;
    },
    
    // nsIObserver + nsIPrefBranchObserver interface implementation
    observe: function(subject, topic, prefName) { 
    },
    
    // nsISupports interface implementation
    QueryInterface: function(iid) {
        if (!iid.equals(Components.interfaces.nsISupports) &&
            !iid.equals(Components.interfaces.nsISupportsWeakReference) &&
            !iid.equals(Components.interfaces.nsIFactory) &&
            !iid.equals(Components.interfaces.nsIObserver))
        {
            dumpToConsole("Cookie consent factory object: QI unknown interface: " + iid + "\n");
            throw Components.results.NS_ERROR_NO_INTERFACE;
        }
	
        return this;
    }
}


function cookieLogConsentInit() {    
    // set preferences to use P3P policy
    var Branch = getPrefService().getBranch("network.cookie.");
    Branch.setIntPref("cookieBehavior",3);

    var compMgr = Components.manager.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    var myCID = Components.ID('{3405B106-AEF2-11DA-B012-B622A1EF5492}');

    compMgr.registerFactory(myCID,COOKIECONSENT_DESCRIPTION,COOKIECONSENT_CONTRACTID,factory);
}

function ForkWindowCookiePermissions(forkWindow,domain,cookiePermissions) {
    this.forkWindow = forkWindow;
    this.domain = domain;
    this.cookiePermissions = cookiePermissions;
}
