// -*- mode: C++; c-basic-offset: 4; fill-column: 75; tab-width: 4; indent-tabs-mode: nil -*-
//
//  **** BEGIN LICENSE BLOCK ****
//  Copyright(c) 2005 Umesh Shankar and Chris Karlof
//  This program is free software; you can redistribute it and/or modify it under
//  the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
//  more details.
//
//  You should have received a copy of the GNU General Public License along with
//  this program; if not, write to the Free Software Foundation, Inc., 59 Temple
//  Place, Suite 330, Boston, MA 02111-1307 USA
//  **** END LICENSE BLOCK ****
//
//
// Library of useful routines for file handling, DOM manipulation, URL processing, etc.
//

const FP_SESSION = 1;
const FP_PERSISTENT = 2;
const TP_SESSION = 4;
const TP_PERSISTENT = 8;
const MAX_POS_PERM = 8;
// These next four are to persist negative decisions
const NO_FP_SESSION = 16;
const NO_FP_PERSISTENT = 32;
const NO_TP_SESSION = 64;
const NO_TP_PERSISTENT = 128;

const POS_TO_NEG_PERMS = 4; // shift by this much to get from pos<->neg perms

// Has there been a decision (positive or negative) made on the given
// permission
function wasDecision(permSet, posPerm) {
    return (permSet & posPerm) ||
        (permSet & (posPerm << POS_TO_NEG_PERMS) );
};



const STATE_START = Components.interfaces.nsIWebProgressListener.STATE_START;
const STATE_STOP = Components.interfaces.nsIWebProgressListener.STATE_STOP;
const STATE_REDIRECTING = Components.interfaces.nsIWebProgressListener.STATE_REDIRECTING;
const STATE_IS_DOCUMENT = Components.interfaces.nsIWebProgressListener.STATE_IS_DOCUMENT;
const STATE_IS_WINDOW = Components.interfaces.nsIWebProgressListener.STATE_IS_WINDOW;

const NS_OK = Components.results.NS_OK;

// Bits of sample code and helper routines

// Adding the "bind" function wraps method calls as regular functions
// so you can say this.method.bind(this) to bind the object instance to the method.
// Useful for callbacks.
Function.prototype.bind = function( object )
{
    var method = this;

    return function() {
        method.apply( object, null );
    }
}

function addBindTo(method) {
    method.bind = function (object) {
        if ( method.length == 0 ) {
            return function() {
                method.apply( object, null );
            }
        } else {
            return function() {
                // I don't want arguments to take the value from inside addBindTo
                method.apply( object, eval("arguments") );
            }
        };
    }
}

// -------
// Observer helpers
//

// Returns a reference to the Mozilla observer service
function observerService() { 
    return Components.classes["@mozilla.org/observer-service;1"]
        .getService(Components.interfaces.nsIObserverService);
}


// The observing class should implement the nsIObserver interface, in particular the observe() function
function addObserver(callbackClass, topicName, addlData) { 
    observerService().addObserver(callbackClass, topicName, addlData);
}

function removeObserver(callbackClass, topicName) {
    observerService().removeObserver(callbackClass, topicName());
}


// This triggers observe() in all the registered observers
// The 'subject' argument is an object relevant to the notification
function notifyObservers(subject, topicName, addlData) {
    observerService().notifyObservers(subject, topicName, addlData);
}
                
//
// End observer helpers
// --------


// -------
// File-handling
// Mostly taken from http://kb.mozillazine.org/File_IO

function ILocalFile() {
    return Components.classes["@mozilla.org/file/local;1"]
	.createInstance(Components.interfaces.nsILocalFile);
}
 
// Path is in platform-specific syntax       
// Return is an nsIFile
function fileFromPath(path) {
    return ILocalFile().initWithPath(path);
}

// String constants for getSpecialDir
// Once you have the directory, you call dir.append("filename") on it to get a filename object
// ProfD 	 profile directory
// DefProfRt 	user (e.g., /root/.mozilla)
// UChrm 	%profile%/chrome
// DefRt 	%installation%/defaults
// PrfDef 	%installation%/defaults/pref
// ProfDefNoLoc 	%installation%/defaults/profile
// APlugns 	%installation%/plugins
// AChrom 	%installation%/chrome
// ComsD 	%installation%/components
// CurProcD 	installation (usually)
// Home 	OS root (e.g., /root)
// TmpD 	OS tmp (e.g., /tmp)
function getSpecialDir(specialName) {
    return Components.classes["@mozilla.org/file/directory_service;1"]
	.getService(Components.interfaces.nsIProperties)
	.get(specialName, Components.interfaces.nsIFile);
}


// File is an nsIFile; the return is an nsIFileOutputStream
// The return of this function must be closed using .close() by the caller
function openForWriting(file, append) {
    // file is nsIFile, data is a string
    var foStream = Components.classes["@mozilla.org/network/file-output-stream;1"]
	.createInstance(Components.interfaces.nsIFileOutputStream);
        
    // use 0x02 | 0x10 to open file for appending.
    if (append)
	foStream.init(file, 0x02 | 0x10, 0664, 0); // write, append
    else
	foStream.init(file, 0x02 | 0x08 | 0x20, 0664, 0); // write, create, truncate


    return foStream;
}
        
// End file handling
// --------------


// -----------
// Window management

function mostRecentBrowserWindow() {
    var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"]
	.getService(Components.interfaces.nsIWindowMediator);
    return wm.getMostRecentWindow("navigator:browser");
}

function getWindowWatcher() {
    var ww = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].
        getService(Components.interfaces.nsIWindowWatcher);
    return ww;
}

function htmlDocFromXULWindow(win) {
    if (! win || ! win.document)
        return null;
    return win.document.getElementById("content").contentDocument;
}

function htmlWinFromXULWindow(win) {
    try { 
        return win.document.getElementById("content").contentWindow;
    } catch(e) {
        return null;
    }
}

// THIS FUNCTION ALWAYS RETURNS NULL (see Bugzilla: https://bugzilla.mozilla.org/show_bug.cgi?id=246632)
// // return is an nsIWebBrowserChrome
// function chromeWinFromHTMLWindow(win) {
//     if (! win)
//         return null;
//     return getWindowWatcher().getChromeForWindow(win);
// }

function appContent(win) {
    if (! win || ! win.document)
        return null;
    return win.document.getElementById("appcontent");
}

// End window management
// -----------------


// -------
// Browser element helpers

function parentFormString(element) {
    if ( element.tagName.toLocaleLowerCase() != "input" || 
         element.form == null ) {
        return ""; 
    }
    
    return (element.form.name + "" + getFormIndex(element.form));
}

function getFormIndex(frm) {
    var doc = frm.ownerDocument;
    for (var i in doc.forms) {
        if (doc.forms[i] == frm) {
            return i;
        }
    }
    return -1;
}

// Try to inject some Javascript into an HTML document, run it, and remove it
// this function stolen from Greasemonkey
function runBrowserScript(doc, jscode) {
    var elm = doc.createElement("script");
    elm.appendChild(doc.createTextNode(jscode));
    doc.body.appendChild(elm);
    doc.body.removeChild(elm);
}

// Get the element's document-relative offset
// Returns an associative array { x: offsetX, y: offsetY }
function getDocOffset(element) {	
    var offsetX = element.offsetLeft;
    var offsetY = element.offsetTop;
    var offsetParent = element.offsetParent;
    while ( offsetParent ) {
        offsetX += offsetParent.offsetLeft;
        offsetY += offsetParent.offsetTop;
        offsetParent = offsetParent.offsetParent;
    }
    return { x: offsetX, y: offsetY };
}

//
// Emulate setting focus on an element (including triggering onfocus())
//
function doFocus(element) {
    element.focus();
    var evt = element.ownerDocument.createEvent('HTMLEvents');
    evt.initEvent('focus', false, false);
    element.dispatchEvent(evt);
}

// emulation for click() method:
// the X,Y coords are element-relative offsets; these may be omitted
// modifier keys may also be omitted
// button: 0 = left, 1 = middle, 2 = right
function doClick(withfocus, element, button, clickCount, clickX, clickY, ctrl, alt, shift, meta) {
    if (clickCount == null || clickCount < 0)
	clickCount = 1;

    var clientX, clientY, screenX, screenY;

    var eltDocOffset = getDocOffset(element);

    if (clickX != null && clickY != null) {
	clientX = eltDocOffset["x"] + clickX - element.ownerDocument.defaultView.pageXOffset;
	clientY = eltDocOffset["y"] + clickY - element.ownerDocument.defaultView.pageYOffset;
    } else {
	clientX = 0;
	clientY = 0;
    }

    screenX = clientX + element.ownerDocument.defaultView.screenX;
    screenY = clientY + element.ownerDocument.defaultView.screenY;
    
    if (ctrl == null)
	ctrl = false;

    if (alt == null)
	alt = false;

    if (shift == null)
	shift = false;

    if (meta == null)
	meta = false;

    var evt;

    if (withfocus) {
        doFocus(element);
    }

    evt = element.ownerDocument.createEvent('MouseEvents');
    evt.initMouseEvent('click', true, true, 
		       element.ownerDocument.defaultView, clickCount, 
		       screenX, screenY, clientX, clientY,
		       ctrl, alt, shift, meta,
		       button, null);
    element.dispatchEvent(evt);
}

// Emulation for reading the outerHTML property
// taken from http://webfx.eae.net/dhtml/ieemu/htmlmodel.html:
var _emptyTags = {
    "IMG":   true,
    "BR":    true,
    "INPUT": true,
    "META":  true,
    "LINK":  true,
    "PARAM": true,
    "HR":    true
};

function getOuterHTML(element) {
    // The node might not be a DOM element
    if (typeof(element.attributes) == "undefined" || 
        typeof(element.attributes.length) == "undefined")
        return "";

    var attrs = element.attributes;
    var str = "<" + element.tagName;
    for (var i = 0; i < attrs.length; i++)
        str += " " + attrs[i].name + "=\"" + attrs[i].value + "\"";
        
    if (_emptyTags[element.tagName])
        return str + ">";
        
    return str + ">" + element.innerHTML + "</" + element.tagName + ">";
}

// pairID() encodes an element as a string, sort of like outerHTML, but encoded
// Returns "tagName=value&attr1=value&...&parentFormString=value&innerHTML=value"
// All value components are encoded with encodeURIComponent()
function pairID(element) {
    var ret = "";
    // The node might not be a DOM element
    if (typeof(element.attributes) == "undefined" || 
	! element.attributes || 
        typeof(element.attributes.length) == "undefined")
        return "";

    ret += "tagName=" + encodeURIComponent(element.tagName);

    var attrs = element.attributes;
    for (var i = 0; i < attrs.length; i++) {
        // Getting the value of the href attribute converts it to a relative URL,
        // but we want the original
        if (attrs[i].name == "href") {
            ret += "&" + attrs[i].name + "=" + encodeURIComponent(element.href);
        } else {
            ret += "&" + attrs[i].name + "=" + encodeURIComponent(attrs[i].value);
        }
    }

    ret += "&parentForm=" + parentFormString(element);
        
    if (_emptyTags[element.tagName] || ! element.textContent) 
        ret += "&textContent=";
    else
        ret += "&textContent=" + encodeURIComponent(element.textContent);

    return ret;
}

// nodeIDString() creates a string "{idstring}TagName(id-attribute)" where idstring = encodeURI(pairID(element))
// for the supplied element in an effort to uniquely identify it.
// 
// The idstring portion can later be parsed by parseIDString();
// changes to this function should be reflected in parseIDString as
// well.
function nodeIDString(element) {
    var idString = "{" + encodeURI(pairID(element)) + "}";

    idString += element.nodeName + "(" + (element.id ? element.id : "") + ")";

    //     var parent = element;
    //     while ( parent ) {
    //         idString += parent.nodeName + "(" + parent.id + ")";
    //         parent = parent.parentNode;
    //     }
    return idString;
}


// Returns
// { "tagName" : value, "textContent" : value, attr1: value, ... }
// with the values decoded
function parseIDString(idString) {
    var ret = new Array();
    var tmp = idString.match(new RegExp("^\\{([^{]+)\\}"));
    var attrString = decodeURI(tmp[1]);

    var values = attrString.split(/&/);
    for (var i = 0; i < values.length; i++) {
        tmp = values[i].split(/=/);
        ret[tmp[0]] = decodeURIComponent(tmp[1]);
    }

    return ret;
};

//
// findElementOneDoc():
// Try to find an element in a document with a certain tag type that best
// matches the supplied ID string. 
//
// Here is the return value format:
// { 
//    elt : [element object], // matched element
//    exact : int,        // exact ID string match
//    id    : int,        // ID ""
//    href  : int,        // HREF attribute matched?
//    name  : int,        // name ""
//    type  : int,        // type ""
//    value  : int,        // value ""
//    textContent : int,        // textContent "
// }
// Each value is 0 (not matched), 1 (both null), or 2 (matched)
//
// Optionally, an existing bestMatch may be passed in (in the result
// format) which must be beaten.
//
// Better match quality is determined by
// findElementMatchCompare(). Ties are settled at the end.
//

var _matchOrder = ["exact", "id", "href", "name", "type", "value", "textContent", "parentForm"];
function _findElementMatchCompare(match1, match2, parsedString)
{
    if (match1 == null)
        return match2;
        
    if (match2 == null)
        return match1;

    var i;
    for (var i = 0; i < _matchOrder.length; ++i) {
        if (match1[_matchOrder[i]] > match2[_matchOrder[i]])
            return match1;
        else if (match2[_matchOrder[i]] > match1[_matchOrder[i]])
            return match2;
    }

    return null; // they were "equal"
}

function _findElementMakeMatch(elt, idString, parsedString) {
    var match = new Array();
    match["elt"] = elt;

    for (var i = 0; i < _matchOrder.length; ++i) {
        var a, b;
        if (_matchOrder[i] == "textContent") {
            a = elt.textContent;
            b = parsedString["textContent"];
        } else if (_matchOrder[i] == "exact") {
            a = nodeIDString(elt);
            b = idString;
        } else if (_matchOrder[i] == "parentForm") {
            a = parentFormString;
            b = parsedString["parentForm"];
        } else {
            a = elt.getAttribute(_matchOrder[i]);
            b = parsedString[_matchOrder[i]];
        }
        
        if ((a == null && b == null) ||
            a === "" && b === "")
            match[_matchOrder[i]] = 1;
        else
            match[_matchOrder[i]] = ((a == b) ? 2 : 0);
    }

    return match;
}

function findElementOneDoc(doc, tagName, idString, bestMatch) {
    var matches = doc.getElementsByTagName(tagName);
    var parsedString = parseIDString(idString);
    var i;
    var ties = new Array();
    for (i = 0; i < matches.length; i++) {
        var thisMatch = _findElementMakeMatch(matches.item(i), idString, parsedString);
        var newBest = _findElementMatchCompare(bestMatch, thisMatch, parsedString);
        if (newBest == thisMatch) {
            bestMatch = thisMatch;
            if (ties.length) { ties = new Array(); }
        } else if (! newBest) {
            ties.push(thisMatch);
        }                   
    }

    // Break ties using edit distance on href
    if (ties && parsedString["href"] != null &&
        bestMatch.elt.href && bestMatch["href"] < 2 &&
        bestMatch.textContent == 2) {

        for (i = 0; i < ties.length; ++i) {

            if ( !ties[i].elt.href ) {
                continue;
            } else if ( ties[i].elt.href == bestMatch.elt.href ) {
                continue;
            } else if ( editDistance(ties[i].elt.href, parsedString["href"]) <
                        editDistance(bestMatch.elt.href, parsedString["href"]) ) {
                bestMatch = ties[i];
            }

        }
    }

    return bestMatch;
}

// 
// findElement() is the main interface for the finding functions above
// 
// Try to find an element recursively in a given window
// Returns a match record as from findElementOneDoc()
//
function findElement(win, tagName, idString, bestMatch) {
    bestMatch = findElementOneDoc(win.document, tagName, idString, bestMatch);
    for (var f = 0; f < win.frames.length; f++) {
        bestMatch = findElement(win.frames[f], tagName, idString, bestMatch);
    }

    return bestMatch;
}

//
// Recursive getElementsByTagName
// Returns a javascript array of matches
//
// twoDim is a boolean: if true, the result is an array of arrays, one per document
//                      if false, one big flat array is returned
//
// The last argument is for internal recursion; it is the set of
// matches so far and should be null (or omitted for the initial call)
//
function recGetElementsByTagName(win, tagName, twoDim, matches) {
    function dumpCollIntoArray(coll, ary) {
        if (twoDim) {
            ary[ary.length] = new Array();
            ary = ary[ary.length-1];
        }

        for (var i = 0; i < coll.length; i++)
            ary[ary.length] = coll[i];
    };

    if (! matches)
        matches = new Array();

    dumpCollIntoArray(win.document.getElementsByTagName(tagName), matches);

    for (var f = 0; f < win.frames.length; f++) {
        var doc = win.frames[f].document;
        dumpCollIntoArray(doc.getElementsByTagName(tagName), matches);

        var iframes = doc.getElementsByTagName("IFRAME");
        var j;
        for (j = 0; j < iframes.length; j++)
            recGetElementsByTagName(iframes[j].contentWindow, tagName, twoDim, matches);

        var frames = doc.getElementsByTagName("FRAME");
        var j;
        for (j = 0; j < frames.length; j++)
            recGetElementsByTagName(frames[j].contentWindow, tagName, twoDim, matches);
    }

    return matches;
}

// Insert text into a currently-focused text box
function insertText(aText) {
    try {
        var command = "cmd_insertText";
        var controller = document.commandDispatcher.getControllerForCommand(command);
        if (controller && controller.isCommandEnabled(command)) {
            controller = controller.QueryInterface(Components.interfaces.nsICommandController);
            var params = Components.classes["@mozilla.org/embedcomp/command-params;1"];
            params = params.createInstance(Components.interfaces.nsICommandParams);
            params.setStringValue("state_data", aText);
            controller.doCommandWithParams(command, params);
        }
    }
    catch (e) {
        dump("Can't do cmd_insertText! ");
        dump(e+"\n")
            }
}

// End browser element helpers
// ----------

// ----------
// Form-related functions 
// "formdata" is an array, [0] = name1, [1] = value1, [2] = name2, ...
// The string representation looks like 
//   name1=uriencodedvalue1&name2=uriencodedvalue2...
// The name portion is formName.inputName; for anonymous forms, we use the
// form's index in the list of all document forms.


function saveOneFormData(frm, prefix, formString) { // prefix = form name or form index usually
    var exclude = { "hidden" : true, "submit" : true };
    var i;

    for (i = 0; i < frm.elements.length; i++) {
	if (! exclude[frm.elements[i].type] && frm.elements[i].name)
	    formString += (formString != "" ? "&" : "") + 
		prefix + "." + frm.elements[i].name + "=" + encodeURIComponent(frm.elements[i].value);
    }

    if (prefix) {
        //	formString += (formString != "" ? "&" : "") + prefix + "." + "action=" + encodeURIComponent(frm.action);
        //	formString += (formString != "" ? "&" : "") + prefix + "." + "target=" + encodeURIComponent(frm.target);
    }

    return formString;
};

function saveAllFormData(win) {
    var formString = "";
    var allforms = recGetElementsByTagName(win, "FORM", true);
                  
    for (var i = 0; i < allforms.length; i++) {
        for (var j = 0; j < allforms[i].length; j++) {
            var frm = allforms[i][j];
            formString = saveOneFormData(frm, (frm.name != "" ? frm.name : j),  formString);
        }
    }

    return formString;
}

// 
// tryFormFillingOneDoc does most of the work: every form field in the
// document which matches a supplied one is filled, if the supplied
// value is different from the existing one. Events are sent to the
// filled-in elements: first focus, then effect the change, then a
// change event, then a blur event.
//
//
// Non-matched input fields are left alone.
// 
// Returns the number of forms filled.
//
function tryFormFillingOneDoc(doc, formdata) {
    var numFilled = 0;
    var forms = doc.getElementsByTagName("FORM");
    var i;
    var j;
    var failed;
    for (i = 0; i < forms.length; i++) {
        var filled_some = 0;
        for (j = 0; j < formdata.length; j += 2) {
            var tmp;
            var inputName = formdata[j];

            // If the form has no "name" property, we go by its position in the forms list
            tmp = formdata[j].match(/^([^.]+)\.(.+)/);
            var formName = null;
            if (tmp && tmp[1] != "") {

                // FIXME: perhaps 1) attach a document name to the form or 2) never re-use a form value (easier)

                if (tmp[1] == forms[i].name) {
                    inputName = tmp[2];
                } else if (tmp[1] == i) {
                    inputName = tmp[2];
                } else {
                    continue;
                }

                formName = tmp[1];

            } else {
                dumpToConsole("Error: missing form name in form data", "formdata-debug");
                debugger;
            }

//             if (inputName == "action") {
//                 forms[i].action = decodeURIComponent(formdata[j+1]);
//                 continue;
//             }
            
//             if (inputName == "target") {
//                 forms[i].target = decodeURIComponent(formdata[j+1]);
//                 continue;
//             }
		
            if (! forms[i].elements.namedItem(inputName))
                continue;
                         
            var elt = forms[i].elements.namedItem(inputName);

            var oldvalue = elt.value;
            var newvalue = decodeURIComponent(formdata[j+1]);

            if (oldvalue != newvalue) {
                //                alert("changing " + formName + "." + elt.name + " from " + oldvalue + " to " + newvalue);
                
                // FIXME: Should we always send the focus event? What about blur()?
                var evt;
            
                elt.focus();
                evt = elt.ownerDocument.createEvent('HTMLEvents');
                evt.initEvent('focus', false, false);
                elt.dispatchEvent(evt);

                elt.value = newvalue;

                evt = elt.ownerDocument.createEvent('HTMLEvents');
                evt.initEvent('change', true, false);
                elt.dispatchEvent(evt);

                
                elt.blur();
                evt = elt.ownerDocument.createEvent('HTMLEvents');
                evt.initEvent('blur', false, false);
                elt.dispatchEvent(evt);                
            }

            filled_some = 1;
        }

        numFilled += filled_some;
    }   
    return numFilled;
};

// Tries to fill in forms for the document in the window, and, if it
// is a frameset, each document in the frameset. This does recurse
// into the documents, themselves, for subframes, IFRAMES, etc.
// Returns the number of forms filled.
function tryFormFilling(win, formdata) {
    var numFilled = 0;
    numFilled += tryFormFillingOneDoc(win.document, formdata);
    for (var f = 0; f < win.frames.length; f++) {
        numFilled += tryFormFilling(win.frames[f], formdata);
    }

    return numFilled;
}

// End form-related functions
// ----------





// ---------
// URL Manipulation

function isChromeURL(url) {
    return (/^chrome:/.test(url) ||
	    /^xul:/.test(url) ||
	    /\.xul$/.test(url) ||
	    /^jar:/.test(url) );
}

// Extracts the hostname part of a URL (the part after the protocol
// and before the port number or next slash
function getURLDomain(arg_url) {
    var url = new URL(arg_url);
    return url.getHost();
}

// FIXME: This should be updated to handle general "effective-TLDs"
// Given a hostname, return the last two components
var knownDomains = /com|net|org|edu|mil|gov|int/;
function getTailURLDomain(domain) {
    if ( ! domain ) { return null; }

    var extraLevel = false;

    var tmp = domain.match(/\.([^\/\.]+)$/);
    if ( tmp ) {
        if ( ! tmp[1].match(knownDomains) ) {
            extraLevel = true;
        }
    } else {        
        return domain; // no last component! (or at least no dot)
    }    

    if ( extraLevel ) {
        tmp = domain.match(/\.([^\/\.]+\.[^\/\.]+\.[^\/\.]+)$/);
    } else {
        tmp = domain.match(/\.([^\/\.]+\.[^\/\.]+)$/);
    }


    if(tmp)
        return tmp[1];
    else
        return domain;
}


// Return the path of a URL (the part following the hostname/port
// number, including an initial slash)
function getURLPath(arg_url) {
    var url = new URL(arg_url);
    return url.getPath();
}


function createURI() {
    return Components.classes["@mozilla.org/network/standard-url;1"].createInstance(Components.interfaces.nsIURI);
}


function equalsURL(url1, url2) {
    if (url1 == null || url2 == null)
	return false;

    return (canonicalizeURL(url1) == canonicalizeURL(url2));
}


function canonicalizeURL(arg_url) {
    if ( arg_url.match(/^keyword/) ) {
        return arg_url;
    }

    var url = new URL(arg_url);
    if (url.getPort() < 0) {
	if (url.getProtocol() == "http")
	    url.setPort(80);
	else if (url.getProtocol() == "https")
	    url.setPort(443);
    }
    
    if (url.getPath() && url.getPath().slice(-1) == "/") 
	url.setPath(url.getPath().slice(0,-1));

    return url.toString();
}

function parseRedirect(redirect,arg_url) {
    if(redirect.charAt(0) == "/") {
	var url = new URL(arg_url);
	url.setFile(redirect);
	return canonicalizeURL(url.toString());
    } else {
	return canonicalizeURL(redirect);
    }
}



// End URL Manipulation
//----------


// ----------
// Cookie-management functions


function getPermissionManager() {
    return Components.classes["@mozilla.org/permissionmanager;1"].getService().QueryInterface(Components.interfaces.nsIPermissionManager);
}

function getCookieService() {
    return Components.classes["@mozilla.org/cookieService;1"] 
        .getService(Components.interfaces.nsICookieService);
}

function addCookie(cookieString,url) {
  var isSession = 1;
  
  var cookieValues = cookieString.split(";");

  var urlDomain = url.match(/\//) ? getURLDomain(url) : url;
  var cookieDomain = null;
  var isSecure = false;

  for(var i=1;i<cookieValues.length;i++) {
    var tmp = cookieValues[i].match(/\s*([^=]+)=(.*)/);
    if(tmp) {
      switch(tmp[1]) {
      case "expires": isSession = 0; break;
          // FIXME: need to implement domain checking to make sure cookie is valid
      case "domain" : cookieDomain = tmp[2]; break; 
      case "secure" : isSecure = true; break;
      }
    }
  }

  var uri = createURI();

  // if cookie specifies domain, for now we allow all cookies of this type from that domain
  // if no domain, use domain of url
  if(cookieDomain) {
    if(cookieDomain.charAt(0) == ".")
      cookieDomain = cookieDomain.substring(1,cookieDomain.length);
    uri.spec = cookieDomain;
  }
  else
    uri.spec = urlDomain;

  // Apparently this has to look like a real URL
  uri.spec = "http" + (isSecure ? "s" : "") + "://" + uri.spec;

  // add cookie to database
  var rv = getCookieService().setCookieString(uri,null,cookieString,null);
  if (! rv) {
      dumpToConsole("Error adding cookie " + cookieString + "for URL " + url);
      debugger;
  }
}

function getCookieManager() {
    return Components.classes["@mozilla.org/cookiemanager;1"] 
        .getService(Components.interfaces.nsICookieManager2);
}

function addCookieObject(cookieObj) {

    var cookieMgr = getCookieManager();
    var addDot = (cookieObj.isDomain && cookieObj.host.charAt(0) != '.') ? "." : "";
    cookieMgr.add( addDot + cookieObj.host, cookieObj.path, 
                   cookieObj.name, cookieObj.value, 
                   cookieObj.isSecure, cookieObj.isSession, cookieObj.expiry );
}


/* function getCookieDomain(cookie,url) { */
/*   var cookieValues = cookie.split(";"); */
/*   var cookieDomain = getURLDomain(url); */
/*   for(var i=1;i<cookieValues.length;i++) { */
/*     tmp = cookieValues[i].match(/\s*([^=]+)=(.*)/); */
/*     if(tmp && tmp[1] == "domain") { */
/*       cookieDomain = tmp[2]; break; // need to implement domain checking to make sure cookie is valid */
/*     } */
/*   } */
/*   return cookieDomain; */
/* } */

function getCookieName(cookieString) {
    var tmp = cookieString.match(/^([^=]*)=.*/);
    if(tmp) {
	dumpToConsole("Cookie name from" + cookieString + " is " + tmp[1]);
	return tmp[1];
    }
    else {
	dumpToConsole("Error: could not find NAME in " + cookie);
	return null;
    }
}

/* function getCookiePath(cookie) { */
/*   var tmp = cookie.match(/;\s*path=[^;\s]*\/); */
/*   if(tmp) */
/*     return tmp[1]; */
/*   else { */
/*     dumpToConsole("Error: could not find path in " + cookie); */
/*     return null; */
/*   } */
/* } */

function classifyCookie(cookie,documentDomain) {
    if(cookie.host.indexOf(getTailURLDomain(documentDomain)) == -1) {
	if(cookie.isSession)
	    return TP_SESSION;
	else
	    return TP_PERSISTENT;
    } else {
	if(cookie.isSession)
	    return FP_SESSION;
	else
	    return FP_PERSISTENT;
    }
}

function parseCookiePermissionsToString(intPerms) {
    if ( intPerms == null ) {
        return "";
    }
    
    var out = "";
    if(intPerms & FP_SESSION)
        out += "FP_SESSION ";
    if(intPerms & FP_PERSISTENT)
        out += "FP_PERSISTENT ";
    if(intPerms & TP_SESSION)
        out += "TP_SESSION ";
    if(intPerms & TP_PERSISTENT)
        out += "TP_PERSISTENT ";
    if(intPerms & NO_FP_SESSION)
        out += "NO_FP_SESSION ";
    if(intPerms & NO_FP_PERSISTENT)
        out += "NO_FP_PERSISTENT ";
    if(intPerms & NO_TP_SESSION)
        out += "NO_TP_SESSION ";
    if(intPerms & NO_TP_PERSISTENT)
        out += "NO_TP_PERSISTENT ";
    if(out != "") {
        out = out.substring(0,out.length-1);
    }
    return out;
}

function parseCookiePermissionsToInt(strPerms) {
    if ( strPerms == null ) {
        return 0;
    }
    var tmp = strPerms.split(" ");
    var intPerms = 0;
    for(i in tmp) {
        if(tmp[i] == "FP_SESSION")
            intPerms |= FP_SESSION;
        else if(tmp[i] == "FP_PERSISTENT")
            intPerms |= FP_PERSISTENT;
        else if(tmp[i] == "TP_SESSION")
            intPerms |= TP_SESSION;
        else if(tmp[i] == "TP_PERSISTENT")
            intPerms |= TP_PERSISTENT;
        else if(tmp[i] == "NO_FP_SESSION")
            intPerms |= NO_FP_SESSION;
        else if(tmp[i] == "NO_FP_PERSISTENT")
            intPerms |= NO_FP_PERSISTENT;
        else if(tmp[i] == "NO_TP_SESSION")
            intPerms |= NO_TP_SESSION;
        else if(tmp[i] == "NO_TP_PERSISTENT")
            intPerms |= NO_TP_PERSISTENT;
        else
            dumpToConsole("Error in parseCookiePermissionsToInt(): perm string =" + tmp[i] + " i=" + i);
    }
    return intPerms;
}

function sanityCheckIntCookiePermissions(intPerms) {
    if ( ( intPerms & FP_SESSION && intPerms & NO_FP_SESSION ) ||
         ( intPerms & FP_PERSISTENT && intPerms & NO_FP_PERSISTENT ) ||
         ( intPerms & TP_SESSION && intPerms & NO_TP_SESSION ) ||
         ( intPerms & TP_PERSISTENT && intPerms & NO_TP_PERSISTENT ) )
        return false;

    return true;
}

function Cookie(nsicookie) {
    this.domain = nsicookie.host;
    this.expires = nsicookie.expires;
    this.host = nsicookie.host;
    this.isDomain = nsicookie.isDomain;
    this.isSecure = nsicookie.isSecure;
    this.name = nsicookie.name;
    this.path = nsicookie.path;
    this.policy = nsicookie.policy;
    this.status = nsicookie.status;
    this.value = nsicookie.value;
    this.expiry = nsicookie.expiry;
    this.isSession = nsicookie.isSession;
    this.rawHost = nsicookie.rawHost;

    this.add = function() {
        var cookieManager = getCookieManager();

        var rv = cookieManager.add(this.domain, this.path, this.name, this.value,
                                   this.isSecure, this.isSession, this.expiry);

        if ( rv ) {
            dumpToConsole("Error adding cookie object in Cookie.add()");
            return false;
        }
        return true;
    };

    this.remove = function() {
        var cookieManager = getCookieManager();

        var rv = cookieManager.remove(this.host, this.name, this.path, false);
        if ( rv ) {
            dumpToConsole("Error removing cookie object in Cookie.remove()");
            return false;
        }
        return true;
    };
}


function cookieFromString(cookiestring,url) {
    var cookieValues = cookiestring.split(";");
    this.domain = getURLDomain(url);
    this.isSession = true;
    this.path = getURLPath(url);
    this.isSecure = false;
    this.isDomain = false;
  
    tmp = cookieValues[0].match(/^\s*([^=]+)=(.*)/);
    if(!tmp) {
	dumpToConsole("Error in creating cookie " + cookiestring + " for URL " + url);
    }

    this.name = tmp[1];
    this.value = tmp[2];
    
    for(var i=1;i<cookieValues.length;i++) {
	tmp = cookieValues[i].match(/\s*([^=]+)=?(.*)/);
	if(tmp) {
	    switch(tmp[1]) {
	    case "expires": this.isSession = false; tmp[2] = tmp[2].replace(/-/g," "); this.expires = parseInt(Date.parse(tmp[2])); break;
	    case "domain" : 
            this.domain = tmp[2]; 
            this.isDomain = true;
            break; // need to implement domain checking to make sure cookie is valid
	    case "path"   : this.path = tmp[2]; break;
	    case "secure" : this.isSecure = true; break;
	    }
	} else {
	    dumpToConsole("Error in matching cookie " + cookieValues[i]);
	}
    } 
}

function cookieToString(cookie) {
    var date = new Date(cookie.expires*1000);
    var ret = cookie.name + "=" + cookie.value + "; ";
    if(!cookie.isSession)
	ret += "expires=" + date.toGMTString() + "; ";

    ret += "domain=" + cookie.host + "; ";

    if(cookie.isSecure)
	ret += "path=" + cookie.path + "; secure";
    else
	ret += "path=" + cookie.path;
    
    if(cookie.serverCert) {
        var certdb = Components.classes["@mozilla.org/security/x509certdb;1"].getService(Components.interfaces.nsIX509CertDB);
        var cert = certdb.constructX509FromBase64(cookie.serverCert);
        ret += "\n" + getCertInfo(cert);
    }
    return ret;
}

function setOfCookiesToString(cookies) {
    var ret = "";
    var i = 0;
    for (i in cookies) {
	ret += cookieToString(cookies[i]) + "\n";
    }
    return ret;
}

// End Cookie-management functions
// ----------


//
//  Window and channel manipulation
//

function getBrowserFromXULWindow(xulWin) {
    return xulWin.document.getElementById("content");
}

function getWebProgressFromChannel(oHttp) {
    var progress = null;
    var result = null;
    oHttp.QueryInterface(Components.interfaces.nsIHttpChannel);
    var name = oHttp.URI.asciiSpec;
    try {
        if( oHttp.notificationCallbacks ) {
            if ( oHttp.notificationCallbacks instanceof XMLHttpRequest ) {
                oHttp.notificationCallbacks.channel.loadGroup.notificationCallbacks.QueryInterface(Components.interfaces.nsIInterfaceRequestor);
                progress = oHttp.notificationCallbacks.channel.loadGroup.
                    notificationCallbacks.getInterface(Components.interfaces.nsIWebProgress,result);
            } else {
                oHttp.notificationCallbacks.QueryInterface(Components.interfaces.nsIInterfaceRequestor);
                progress = oHttp.notificationCallbacks.getInterface(Components.interfaces.nsIWebProgress,result);
            }
        } else if ( oHttp.loadGroup && oHttp.loadGroup.notificationCallbacks ) {
            oHttp.loadGroup.notificationCallbacks.QueryInterface(Components.interfaces.nsIInterfaceRequestor);
            progress = oHttp.loadGroup.notificationCallbacks.getInterface(Components.interfaces.nsIWebProgress,result);
        } else {
            dumpToConsole("No notification callbacks for " + name, "domwindow-debug");
        }
    } catch(ex) {
        debugger;
        dumpToConsole("No notification callbacks (exception) for " + name , "domwindow-debug");    
    }
    return progress;
}

function getDOMWindowFromChannel(oHttp) {
    var progress = getWebProgressFromChannel(oHttp);
    if(progress) {
        return progress.DOMWindow;
    } else {
        return null;
    }
}

//Taken from Venkman
function getNSIXULWindowFromDOMORChromeWindow (win)
{
    var rv;
    //dd ("getXULWindowFromWindow: before: getInterface is " + win.getInterface);
    try
        {
            var requestor = win.QueryInterface(Components.interfaces.nsIInterfaceRequestor);
            var nav = requestor.getInterface(Components.interfaces.nsIWebNavigation);
            var dsti = nav.QueryInterface(Components.interfaces.nsIDocShellTreeItem);
            var owner = dsti.treeOwner;
            requestor = owner.QueryInterface(Components.interfaces.nsIInterfaceRequestor);
            rv = requestor.getInterface(Components.interfaces.nsIXULWindow);
        }
    catch (ex)
        {
            rv = null;
            //dd ("not a nsIXULWindow: " + formatException(ex));
            /* ignore no-interface exception */
        }
 
    //dd ("getXULWindowFromWindow: after: getInterface is " + win.getInterface);
    return rv;
}

//taken from venkman
function getNSIBaseWindowFromDOMOrChomeWindow (win)
{
    var rv;
    try
        {
            var requestor = win.QueryInterface(Components.interfaces.nsIInterfaceRequestor);
            var nav = requestor.getInterface(Components.interfaces.nsIWebNavigation);
            var dsti = nav.QueryInterface(Components.interfaces.nsIDocShellTreeItem);
            var owner = dsti.treeOwner;
            requestor = owner.QueryInterface(Components.interfaces.nsIInterfaceRequestor);
            rv = requestor.getInterface(Components.interfaces.nsIBaseWindow);
        }
    catch (ex)
        {
            rv = null;
        }
 
    return rv;
}

function hideDOMOrChromeWindow(win) {
    getNSIBaseWindowFromDOMOrChomeWindow(win).visibility = false;
}

function unhideDOMOrChromeWindow(win) {
    getNSIBaseWindowFromDOMOrChomeWindow(win).visibility = true;
}

function getServerCertFromChannel(oHttp) {
    oHttp.QueryInterface(Components.interfaces.nsIHttpChannel);
    var sp = oHttp.securityInfo;
    if(sp) {
        sp.QueryInterface(Components.interfaces.nsISSLStatusProvider);
        var status = sp.SSLStatus;
        if(status) {
            dumpToConsole(oHttp.URI.spec);
            status.QueryInterface(Components.interfaces.nsISSLStatus);
            return status.serverCert;
        }
    }
    return null;
}

function isElementOfLoadGroup(url,loadGroup) {
    var iterator = loadGroup.requests;
    while(iterator.hasMoreElements()) {
	var current = iterator.getNext();
	current.QueryInterface(Components.interfaces.nsIRequest);
	if(current.name == url) {
	    return true;
	}
    }
    return false;
}

function loadGroupToString(loadGroup) {
    var iterator = loadGroup.requests;
    var out = "LOADGROUP: " + loadGroup.defaultLoadRequest + "\n";
    while(iterator.hasMoreElements()) {
	var current = iterator.getNext();
	current.QueryInterface(Components.interfaces.nsIRequest);
	out += current.name + "\n";
    }
    return out;
}




//
// Preferences
//

var NS_PREFBRANCH_PREFCHANGE_TOPIC_ID = "nsPref:changed";

function getPrefService() {
    return Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
}

function getCookielogBranch() {
    return getPrefService().getBranch("cookielog.").QueryInterface(Components.interfaces.nsIPrefBranch2);
}





//
// Debugging
//

function internalError(aMessage) {
    dumpToConsole("INTERNAL ERROR: " + aMessage);
    alert("DOPPELGANGER INTERNAL ERROR: Open the debugger if you are a developer, then click OK.\n" 
          + aMessage);
    debugger;
}

function getConsoleService() {
    return Components.classes['@mozilla.org/consoleservice;1'].getService(Components.interfaces.nsIConsoleService);
}

function dumpToConsole(aMessage,type) {
    var consoleService = getConsoleService();
    if(type) {
        var prefObj = getPrefService();
        var Branch = prefObj.getBranch("cookielog.");
        if(Branch.prefHasUserValue(type)) {
            if(Branch.getBoolPref(type)) {
                consoleService.logStringMessage("Doppelganger: " + aMessage);
            }
        } else {
            consoleService.logStringMessage("CookieLog (WARNING DEBUG TYPE " + type + " NOT KNOWN): " + aMessage);
        }
    } else {
	consoleService.logStringMessage("Doppelganger: " + aMessage);
    }
}

// The 'file' argument is an open nsIFileOutputStream
function writeConsoleToFile(foStream) {
    var consoleService = getConsoleService();
    var out = {};
    consoleService.getMessageArray(out, {});
    var messages = out.value;
    for ( var i in messages ) {
        foStream.write(messages[i].message + "\n", messages[i].message.length + 1);
    }
}

function dump_obj(prefix, obj) {
    var i;
    try {
	if (obj) {
	    for (i in obj) {
		try {
		    if (! obj[i]) {
			dump(prefix + "[UNDEF] " + i + "\n");
		    }
		    else if (typeof(obj[i]) == "function") {
			dump(prefix + "[" + typeof(obj[i]) + "] " + i + "\n");
		    }
		    else if (typeof(obj[i]) == "object") {
			dump(prefix + "[" + typeof(obj[i]) + "] " + i + " = " + obj[i] + "\n");
		    }
		    else {
			dump(prefix + "[" + typeof(obj[i]) + "] " + i + " = " + obj[i] + "\n");
		    }
		}
		catch (exc) {
		    alert("ERROR on " + i + ": " + exc);
		}
	    }
	}
	else {
	    dump(prefix + "OBJECT IS UNDEFINED\n");
	}
    }
    catch (exc2) {
	dump(prefix + "ANY ERROR\n");
    }
}




//
// String processing
//

function chomp(s) {
    var tmp = s.match(/(.*)\r?\n/);
    if(tmp) {
        return tmp[1];
    } else {
        return s;
    }
}

// An implementation of the Levenshtein edit distance algorithm
// by Lukasz Stilger
function edMinimum(a, b, c) {
    var mi;
    mi = a;
    if (b < mi)
        mi = b;
    if (c < mi)
        mi = c;
    return mi;
}

function editDistance(s, t) {
    var d = new Array();
    var n; // length of s
    var m; // length of t
    var i; // iterates through s
    var j; // iterates through t
    var s_i; // ith character of s
    var t_j; // jth character of t
    var cost; // cost

//     var dG = new Array();		

    // Step 1
    n = s.length;
    m = t.length;
    if (n == 0) {
        return m;
    }

    if (m == 0) {
        return n;
    }
			
    //inicjacja tablicy dwu-wymiarowej w Javascript	
    for(i=0; i<=n; i++)
        d[i] = new Array();


    // Step 2
    for (i = 0; i <= n; i++) {
        d[i][0] = i;
    }

    for (j = 0; j <= m; j++) {
        d[0][j] = j;
    }

    // Step 3
    for (i = 1; i <= n; i++) {

        s_i = s.charAt(i - 1);

				
        // Step 4
        for (j = 1; j <= m; j++) {

            t_j = t.charAt(j - 1);

            // Step 5
            if (s_i == t_j) {
                cost = 0;
            }
            else {
                cost = 1;
            }

            // Step 6
            d[i][j] = edMinimum (d[i-1][j]+1, d[i][j-1]+1, d[i-1][j-1] + cost);
        }

    }
			
//     //przepisanie do tablicy globalnej
//     for(i=1; i<=n; i++) {
//         dG[i] = new Array();
//         for(j=1; j<=m; j++)
//             dG[i][j] = d[i][j];
//     }

    // Step 7
    return d[n][m];
}
