//
// P3P parsing functions taken from PrivacyFox by Fahd Arshad (Copyright 2005)
// Small modifications made so they can be called from other code
//  -- Umesh
//


const notWhitespace = /\S/;


// Returns the URL of the P3P policy
function LoadRefFile(xmlDoc, protoAndHost) {

 
  // ASSMUE: There should only ever by one policy-references tag in a ref file. 
  // ASSUME: Reference file's XML tags are in uppercase.
  // ASSUME: policy ref file exists in well-known loacation (no graceful exit if it doesn't).
  var obj = xmlDoc.getElementsByTagName("POLICY-REFERENCES")[0];
  //	Notify("childNodes length before whitespace removal is " + obj.childNodes.length);
  // remove whitespaces from DOM so that they don't count as nodes
  if (!obj) {
      //    alert("Privacy Fox error:\nI can't seem to parse this site's policy reference file. Sorry!");
    return null;
  }

  for (i=0; i<obj.childNodes.length; i++) {
    if ((obj.childNodes[i].nodeType == 3) && (!notWhitespace.test(obj.childNodes[i].nodeValue))) {
      obj.removeChild(obj.childNodes[i]);
      i--;
    }
  }
  //	Notify("ChildNodes length after whitespace removal is " + obj.childNodes.length);

  //  for (i=0; i<obj.childNodes.length; i++) {	
  // var str = obj.childNodes[i].firstChild.nodeValue;
  // Notify("The nodeValue of Policy-references' first child (policy-ref) is: " + str);
  // var policy = obj.childNodes[i].getAttribute("about");
  var polRefObj = obj.getElementsByTagName("POLICY-REF")[0]; // The first POLICY tag
  if (!polRefObj) {
      //     alert("Privacy Fox error:\nI can't seem to parse this site's policy reference file. Sorry!");
     return null;
  }   
  var policy = polRefObj.getAttribute("about");
  if (!policy) {
      //     alert("Privacy Fox error:\nI can't seem to parse this site's policy reference file. Sorry!");
     return null;
  }
  // policy name could be absolute or relative. Handle the relative (bravenet.net)
  var regexpURI = /http:/;
  if (regexpURI.test(policy)) {
  }
  else {
    var regexpRel = /^\//;      
    var regexpInc = /^#/;
    if (regexpRel.test(policy)) {
      policy = protoAndHost + policy;      
    }
    else if (regexpInc.test(policy)) {
      policy = RefURI + policy;      
    }
    else {
	//      alert("Privacy Fox error:\nI can't seem to parse this site's policy reference file. Sorry!");
      return null;
    }
  }

  return policy;
}

/***************************************************************************
  ParsePolicyFile: parses the policy file into an hash called Store{}. Calls Translate() at the end.
****************************************************************************/
function Notify(msg) {
    dumpToConsole(msg, "p3p-debug");
}


// Returns a bunch of key-value pairs
function ParsePolicyFile(xmlDoc){
    var Store = new Array;

  Notify("ParsePolicyFile called.");
  // BUILD: EXPIRY tag is checked
  var obj = xmlDoc.getElementsByTagName("POLICIES")[0]; // First policies tag; there's only supposed to be 1
  if (obj) {
    Notify(" Found the POLICIES tag..");    
    for (i=0; i<obj.childNodes.length; i++) { // remove whitespace so it doesn't get treated as nodes
      if ((obj.childNodes[i].nodeType == 3) && (!notWhitespace.test(obj.childNodes[i].nodeValue))) {
	obj.removeChild(obj.childNodes[i]);
	i--;
      }
    }
  }
  else {
    Notify("<font color=\"red\">My call to getElementsByTagName failed for tag POLICIES.</font>");
    //    alert("Privacy Fox error:\nI can't seem to parse this site's policy. Sorry!");
    return -1;
  }
  // BUILD: Multiple <POLICY> tags to be handled
  //if (obj.childNodes.length > 1) { // if there are multiple <POLICY>  tags
  // Notify("There are multiple <POLICY>s in this policy file.");
  // expiry or policy tag can count here, so the above line is wrong.
  //}
  // BUILD: EXPIRY attribute is checked.
  // ASSMUE: DATASCHEMA and lang attributes can be safely ignored
  //	var polObj = obj.childNodes[0]; // The first POLICY tag
  var polObj = obj.getElementsByTagName("POLICY")[0]; // The first POLICY tag
  Store.policyName = polObj.getAttribute("name");
  Store.discURI = polObj.getAttribute("discuri");
  Store.optURI = polObj.getAttribute("opturi");
  var out = "The policy";
  if (Store.policyName == null) {
    Notify("WARNING: the policy has no name.");
  }
  else {
    out += " \"" + Store.policyName + "\"";
  }
  if (Store.discURI == null) {
    Notify("WARNING: the discURI attribute is missing.");
    out += " has no human-readable version";
  }
  else {
    out += " has a human-readable version at " + Store.discURI;
  }
  if (Store.optURI != null) {
    out += " and the opt-out information is available at <" + Store.optURI + ">";
  }
  Notify(out + ".");
	
  // Parsing <TEST>
  tstObj = polObj.getElementsByTagName("TEST")[0];
  if (tstObj) {
    Notify("This is a test policy.");
    Store.TEST = 1;
  }
	
  // Parsing <ENTITY>
  entObj = polObj.getElementsByTagName("ENTITY")[0];
  if (entObj) {
    //		Notify("I did find the contact information for this website.");
    var entDGrp = entObj.getElementsByTagName("DATA-GROUP")[0];
    if (entDGrp) {
      for (i=0; i<entDGrp.childNodes.length; i++) {
	if (entDGrp.childNodes[i].nodeType != 3) {
	  var type = entDGrp.childNodes[i].getAttribute("ref");
	  var typeStr = type.substr(1,type.length-1);
	  Store[typeStr] = entDGrp.childNodes[i].firstChild.nodeValue;
	  //					Notify("<" + entDGrp.childNodes[i].tagName + ">" + typeStr + " : " + Store[typeStr] );
	}
      }
    }
    else {
      Notify("I couldn't find the DATA-GROUP tag in ENTITY");
    }	
  }
  else {
    Notify("WARNING: No contact information was found for this policy!");
  }

  // Parse <ACCESS>
  var accObj = polObj.getElementsByTagName("ACCESS")[0];
  if (accObj) {
    if (accObj.childNodes.length > 0) {
      Store.ACCESS = 1;
    }
    for (i = 0 ; i < accObj.childNodes.length; i++) {
      if (accObj.childNodes[i].tagName) {
	Store["ACCESS-" + accObj.childNodes[i].tagName] = 1;
      }
    }
  }
  else {
    Notify("WARNING: I couldn't find an ACCESS tag!");
  }

  // Parse DISPUTES-GROUP. DISPUTES resol. can be of 4 types: service, indep, court, & law
  // Is there always only one DISPUTES-GROUP
  var dispGrpObj = polObj.getElementsByTagName("DISPUTES-GROUP")[0];
  if (!dispGrpObj) {
    Notify("WARNING: I couldn't find an <DISPUTES-GROUP> tag!");
    Store["DISPUTES-GROUP"] = 0;
  }
  else {
    var dispGrpArr = dispGrpObj.childNodes;
    Store["DISPUTES-GROUP"] = 1;
    // Notify("The # of children of dispute-group is " + dispGrpArr.length);
    for (i=0; i<dispGrpArr.length; i++) {
      if (dispGrpArr[i].tagName) {
	var key = "DISPUTES-" + dispGrpArr[i].getAttribute("resolution-type");
	Store[key] = 1;
	// key += "-" + "service";
	Store[key + "-service"] = dispGrpArr[i].getAttribute("service");
	Store[key + "-sDesc"] = dispGrpArr[i].getAttribute("short-description");
	if (dispGrpArr[i].getAttribute("verification")) {
	  Store[key + "-verif"] = dispGrpArr[i].getAttribute("verification");
	}
	var ldObj = dispGrpArr[i].getElementsByTagName("LONG-DESCRIPTION")[0];
	if (ldObj) {
	  // Notify("The long description is [" + ldObj.firstChild.nodeValue + "]");
	  Store[key + "-lDesc"] = ldObj.firstChild.nodeValue;
	}
	var remObj = dispGrpArr[i].getElementsByTagName("REMEDIES")[0];
	if (remObj) {
	  if (remObj.getElementsByTagName("correct")[0]) {
	    // Notify("Correct was available as one of the remedies.");
	    Store[key + "-rem-correct"] = 1;
	  }
	  if (remObj.getElementsByTagName("money")[0]) {
	    // Notify("Money was available as one of the remedies.");
	    Store[key + "-rem-money"] = 1;
	  }
	  if (remObj.getElementsByTagName("law")[0]) {
	    // Notify("Law was available as one of the remedies.");
	    Store[key + "-rem-law"] = 1;
	  }
	}		
	
      } // end if tagname
    } // end for 
  } // end else -> if dispGrpObj
	
  // Get the statements
  var stmtArr = polObj.getElementsByTagName("STATEMENT");
  Notify("The number of statements in this policy is " + stmtArr.length);
  Store["numST"] = stmtArr.length;

  // Parsing individual statements 
  for (s = 0; s < stmtArr.length; s++) {
    Notify("Parsing statement # " + s + ".");
    // var stmtObj = stmtArr[s];

    // Parse CONSEQUENCE (optional)
    var cnsqObj = stmtArr[s].getElementsByTagName("CONSEQUENCE")[0];
    if (cnsqObj) { // CONSEQUENCE is optional
      Store["ST" + s + ".CONSEQUENCE"] = cnsqObj.firstChild.nodeValue;
    }
    else {
      Notify("This statement had no CONSEQUENCE tag (it is optional).");
    }

    // Parse NON-IDENTIFIABLE (optional)
    var nonident = stmtArr[s].getElementsByTagName("NON-IDENTIFIABLE")[0];
    if (nonident) {
      Store["ST" + s + ".NON-IDENTIFIABLE"] = 1;
    }
    //   else { // the rest of the statement matters only if no non-ident claim is made

    // Parse PURPOSE
    var prpsObj = stmtArr[s].getElementsByTagName("PURPOSE")[0];
    if (prpsObj) {
      var prpsArr = prpsObj.childNodes;
      if (prpsArr.length > 0) {
	Store["ST" + s + ".PURPOSE"] = 1;
      }
      for (var i = 0; i < prpsArr.length; i++) {
	if (prpsArr[i].tagName) {
	  //	Notify("Purpose: <" + prpsObjArr[i].tagName + ">");	  
	  var req = prpsArr[i].getAttribute("required");
	  if (req) {
	    Store["ST" + s + ".PURPOSE-" + prpsArr[i].tagName] = req;
	  }
	  else {
	    Store["ST" + s + ".PURPOSE-" + prpsArr[i].tagName] = 1;
	  }
	  if (prpsArr[i].tagName == 'other-purpose') {
	    Store["ST" + s + ".other-purpose"] = prpsArr[i].nodeValue;
	  }
	}
      }
    }
    else {
      Store["ST" + s + ".PURPOSE"] = 0;
    }   

    // Parse RECEPIENT
    var rcptObj = stmtArr[s].getElementsByTagName("RECIPIENT")[0];
    if (rcptObj) {
      var rcptArr = rcptObj.childNodes;
      if ( rcptArr.length > 0) {
	Store["ST" + s + ".RECEPIENT"] = 1;
      }
      for (var i = 0; i < rcptArr.length; i++) {
	if (rcptArr[i].tagName) {
	  //	Notify("Purpose: <" + prpsObjArr[i].tagName + ">");
	  Store["ST" + s + ".RECEPIENT-" + rcptArr[i].tagName] = 1;
	}
      }
    }
    else {
      Store["ST" + s + ".RECEPIENT"] = 0;
    }      
    
    // Parse RETENTION
    var rtnObj = stmtArr[s].getElementsByTagName("RETENTION")[0];
    if (rtnObj) {
      var rtnArr = rtnObj.childNodes;
      if (rtnArr.length > 0) {
	Store["ST" + s + ".RETENTION"] = 1;
      }
      for (var i = 0; i < rtnArr.length; i++) {
	if (rtnArr[i].tagName) {
	  //	Notify("Purpose: <" + prpsObjArr[i].tagName + ">");
	  Store["ST" + s + ".RETENTION-" + rtnArr[i].tagName] = 1;
	}
      }
    }
    else {
      Store["ST" + s + ".RETENTION"] = 0;
    }

    // Parse DATA-GROUP
    var dataGrp = stmtArr[s].getElementsByTagName("DATA-GROUP")[0];
    if (dataGrp) {
      var dataArr = dataGrp.getElementsByTagName("DATA");
      Store["ST" + s + ".numData"] = dataArr.length;
      for (var i = 0; i < dataArr.length; i++) {
	//	Notify("Data ref: <" + dataArr[i].getAttribute("ref") + ">");
	var type = dataArr[i].getAttribute("ref");
	var typeStr = type.substr(1,type.length-1);
	var catObj = dataArr[i].getElementsByTagName("CATEGORIES")[0];
	if (catObj) { // CATEGORIES is optional
	  var catArr = catObj.childNodes;
	}
	var catStr = "";
	if (catArr) {
	  for (var c = 0; c < catArr.length; c++) {
	    if (catArr[c]) {
	      //		Notify("ST" + s + ".DATA-" + i + "." + typeStr + "+" + catArr[c].tagName);
	      catStr += catArr[c].tagName + "+";
	    }
	  }
	}
	if (dataArr[i].getAttribute("optional")) {
	  Store["ST" + s + ".DATA-" + i] = typeStr + "-" + catStr + "=" + "optional";
	}
	else {
	  Store["ST" + s + ".DATA-" + i] = typeStr + "-" + catStr + "=" + "notOptional";
	}
      } // end for data
    } // end if dataGrp
    else {
      Store["ST" + s + ".numData"] = 0;
    }

  } // end for loop throurgh statements

  Notify("From ParsePolicyFile, calling Translate on policy named " + Store.policyName);

  return Store;

}

