// -*- mode: c++; c-basic-offset: 4; fill-column: 75; -*-
/* URL class
 * Copyright 2005 Umesh Shankar <ushankar@cs.berkeley.edu>
 *
 * Originally based on code
 * Copyright (C) 2003 Johan Knngrd, <johan AT kanngard DOT net>
 * http://dev.kanngard.net/
 *
 * However, now completely different.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The GPL is located at: http://www.gnu.org/licenses/gpl.txt
 */

function URL(url){
    var tmp; // for regex matching

    if(url.length==0) { 
	throw "Invalid URL: url";
    }

    this.url=url;

    this.port=-1;

    // Pull off #, ? and path into member variables, and return the remaining prefix
    this.parseOffRefQueryAndPath = function (s) {
	var tmp;

	if ((tmp = s.match(/#(.*)$/))) {
	    this.reference = tmp[1];
	    s = s.replace(/#(.*)$/, "");
	} 

	if ((tmp = s.match(/\?(.*)$/))) {
	    this.query = tmp[1];
	    s = s.replace(/\?(.*)$/, "");
	};

	if ( (tmp = s.match( /\/\/[^\/]*(\/.*$)/ )) ) {
	    this.path = tmp[1];
	    s = s.substr(0, s.length - this.path.length);
	};

	return s;
    }

    this.buildFile = function() {
	this.file = this.path;

	if ( this.query != null && this.query.length > 0 ) 
	    this.file+='?'+this.query;
	if ( this.reference != null && this.reference.length > 0 ) 
	    this.file+='#'+this.reference;	
    };

    /* Returns the port part of this URL, i.e. '8080' in the url 'http://server:8080/' */
    this.getPort = function() {
	return this.port;
    }

    /* Returns the query part of this URL, i.e. 'Open' in the url 'http://server/?Open' */
    this.getQuery = function() {
	return this.query;
    }

    /* Returns the protocol of this URL, i.e. 'http' in the url 'http://server/' */
    this.getProtocol = function() {
	return this.protocol;
    }

    /* Returns the host name of this URL, i.e. 'server.com' in the url 'http://server.com/' */
    this.getHost = function() {
	return this.host;
    }

    /* Returns the user name part of this URL, i.e. 'joe' in the url 'http://joe@server.com/' */
    this.getUserName = function() {
	return this.username;
    }

    /* Returns the password part of this url, i.e. 'secret' in the url 'http://joe:secret@server.com/' */
    this.getPassword = function() {
	return this.password;
    }

    /* Returns the file part of this url, i.e. everything after the host name. */
    this.getFile = function() {
	return this.file;
    }

    /* Returns the reference of this url, i.e. 'bookmark' in the url 'http://server/file.html#bookmark' */
    this.getReference = function() {
	return this.reference;
    }

    /* Returns the file path of this url, i.e. '/dir/file.html' in the url 'http://server/dir/file.html' */
    this.getPath = function() {
	return this.path;
    }

    this.setPort = function(port) {
	this.port = port;
    }

    this.setQuery = function(query) {
	this.query = query;
	this.buildFile();
    }

    this.setProtocol = function(protocol) {
	this.protocol = protocol;
    }

    this.setHost = function(host) {
	this.host = host;
    }

    this.setUserName = function(username) {
	this.username = username;
    }

    this.setPassword = function(password) {
	this.password = password;
    }

    this.setFile = function(file) {
	this.file = file;
	this.parseOffRefQueryAndPath(this.file);
    }

    this.setReference = function(reference) {
	this.reference = reference;
	this.buildFile();
    }

    this.setPath = function(path) {
	this.path = path;
	this.buildFile();
    }



    /* Returns the FIRST matching value to the specified key in the query.
       If the url has a non-value argument, like 'Open' in '?Open&bla=12', this method
       returns the same as the key: 'Open'...
       The url must be correctly encoded, ampersands must encoded as &amp;
       I.e. returns 'value' if the key is 'key' in the url 'http://server/?Open&amp;key=value' */
    this.getArgumentValue = function(key){
	var a=this.getArgumentValues();
	if(a.length<1) return '';
	for(i=0;i<a.length;i++){
	    if(a[i][0]==key) return a[i][1];
	}
	return '';
    }

    /* Returns all key / value pairs in the query as a two dimensional array */
    this.getArgumentValues = function(){
	var a=new Array();
	var b=this.query.split('&');
	var c='';
	if( b.length < 1 ) return a;
	for( var i = 0; i < b.length; i++ ) {
	    c = b[i].split('=');
	    a[i] = new Array(c[0],((c.length==1)?c[0]:c[1]));
	}
	return a;
    }

    this.toString = function(){
	if (this.protocol == "about") {
	    return "about:" + this.file;
	}

	var userpass = this.username;
	if (userpass != null) {
	    userpass += (this.password ? (":" + this.password) : "");
	    userpass += "@";
	}

	return this.protocol + "://" + 
	    (userpass != null ? userpass : "") + 
	    this.host + 
	    (this.port != null ? (":" + this.port) : "") +
	    (this.path ? this.path : "") + 
	    (this.query != null ? ("?" + this.query) : "") + 
	    (this.reference != null ? ("#" + this.reference) : "");
    }

    try{
	if (url.match(/^about:/)) {
	    this.protocol = "about";
	    this.file = this.path = url.substr(6);
	    return;
	}
    } catch (e) {
	debugger;
    }

    var prefix = this.parseOffRefQueryAndPath(url);

    tmp = prefix.match(/^(.*):\/\//);
    if (tmp) {
	this.protocol = tmp[1];
	prefix = prefix.replace(/^(.*):\/\//, "");
    } else {
	throw "No protocol component in URL " + url;
    }

    if (prefix.indexOf("@") >= 0) {
	tmp = prefix.split("@");
	var userAndPass = tmp[0];
	prefix = tmp[1];

	if (userAndPass.indexOf(":") >= 0) {
	    tmp = userAndPass.split(":");
	    this.username = tmp[0];
	    this.password = tmp[1];
	} else {
	    this.username = userAndPass;
	    this.password = "";
	}
    } else {
	this.username = null;
	this.password = null;
    }

    tmp = prefix.match (/^([-\w.]+)(:\d+)?$/);
    if (tmp) {
	this.host = tmp[1];
	if (tmp[2])
	    this.port = tmp[2].substr(1);
	else
	    this.port = null;
    } else {
	throw "No host component in URL " + url;
    }
    
    this.buildFile();

}
